""" API testcases for the "/user/" endpoint """
from .base import MyApiTestCase
from privacyidea.lib.resolver import save_resolver, get_resolver_object
from privacyidea.lib.realm import set_realm
from privacyidea.lib.user import User
from privacyidea.lib.token import init_token, remove_token
from urllib.parse import urlencode, quote

PWFILE = "tests/testdata/passwd"


class APIUsersTestCase(MyApiTestCase):

    parameters = {'Driver': 'sqlite',
                  'Server': '/tests/testdata/',
                  'Database': "testuser-api.sqlite",
                  'Table': 'users',
                  'Encoding': 'utf8',
                  'Map': '{ "username": "username", \
                    "userid" : "id", \
                    "email" : "email", \
                    "surname" : "name", \
                    "givenname" : "givenname", \
                    "password" : "password", \
                    "phone": "phone", \
                    "mobile": "mobile"}'
                  }

    def _create_user_wordy(self):
        """
        This creates a user "wordy" in the realm "sqlrealm"
        """
        realm = "sqlrealm"
        resolver = "SQL1"
        parameters = self.parameters
        parameters["resolver"] = resolver
        parameters["type"] = "sqlresolver"

        rid = save_resolver(parameters)
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(realm, [{'name': resolver}])
        self.assertEqual(len(failed), 0)
        self.assertEqual(len(added), 1)
        with self.app.test_request_context('/user/',
                                           method='POST',
                                           data={"user": "wordy",
                                                 "resolver": resolver,
                                                 "surname": "zappa",
                                                 "givenname": "frank",
                                                 "email": "f@z.com",
                                                 "phone": "12345",
                                                 "mobile": "12345",
                                                 "password": "12345"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value") > 6, result.get("value"))

    def _get_wordy_auth_token(self, password="12345"):
        """
        User wordy logs in and gets an auth-token: self.wordy_auth_token
        :return:
        """
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "wordy@{0!s}".format("sqlrealm"),
                                                 "password": password}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            self.wordy_auth_token = result.get("value").get("token")
            # check that this is a user
            role = result.get("value").get("role")
            self.assertEqual("user", role, result)

    def test_00_get_empty_users(self):
        with self.app.test_request_context('/user/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            self.assertTrue(res.json["result"]["status"], res.json)
            self.assertEqual([], res.json["result"]["value"], res.json)

    def test_01_get_passwd_user(self):
        # create resolver
        with self.app.test_request_context('/resolver/r1',
                                           json={"resolver": "r1",
                                                 "type": "passwdresolver",
                                                 "fileName": PWFILE},
                                           method='POST',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            self.assertEqual(res.json['result']['value'], 1, res.json)

        # create realm
        realm = "realm1"
        resolvers = "r1, r2"
        with self.app.test_request_context('/realm/{0!s}'.format(realm),
                                           data={"resolvers": resolvers},
                                           method='POST',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json
            value = result.get("result").get("value")
            self.assertIn('r1', value["added"], result)
            self.assertIn('r2', value["failed"], result)

        # get user list
        with self.app.test_request_context('/user/',
                                           query_string=urlencode({"realm": realm}),
                                           method='GET',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json
            value = result.get("result").get("value")
            unames = [x.get('username') for x in value]
            self.assertIn("cornelius", unames, value)
            self.assertIn("corny", unames, value)
            # Check that there is no password entry in the results
            self.assertTrue(all(["password" not in x for x in value]), value)

        # get user list with search dict
        with self.app.test_request_context('/user/',
                                           query_string=urlencode({"username": "cornelius"}),
                                           method='GET',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json
            value = result.get("result").get("value")
            # Check that there is no password entry in the result
            self.assertNotIn("password", value, result)
            unames = [x.get('username') for x in value]
            self.assertIn("cornelius", unames, value)
            self.assertNotIn("corny", unames, value)

        # Get user with a non-existing realm
        with self.app.test_request_context('/user/',
                                           query_string=urlencode({"realm": "non_existing"}),
                                           method='GET',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json
            value = result.get("result").get("value")
            unames = [x.get('username') for x in value]
            self.assertNotIn("cornelius", unames, value)
            self.assertNotIn("corny", unames, value)

    def test_02_create_update_delete_user(self):
        realm = "sqlrealm"
        resolver = "SQL1"
        parameters = self.parameters
        parameters["resolver"] = resolver
        parameters["type"] = "sqlresolver"

        rid = save_resolver(parameters)
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(realm, [{'name': resolver}])
        self.assertEqual(len(failed), 0)
        self.assertEqual(len(added), 1)

        # CREATE a user
        self._create_user_wordy()

        # Get users
        with self.app.test_request_context('/user/',
                                           method='GET',
                                           query_string=urlencode(
                                               {"username": "wordy"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"))
            self.assertEqual(result.get("value")[0].get("username"), "wordy")

        # Update by administrator. Set the password to "passwort"
        with self.app.test_request_context('/user/',
                                           method='PUT',
                                           query_string=urlencode(
                                               {"user": "wordy",
                                                "resolver": resolver,
                                                "password": "passwort"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

        # Get user authentication and update by user.
        self._get_wordy_auth_token("passwort")
        wordy_auth_token = self.wordy_auth_token

        # Even if a user specifies another username, the username is
        # overwritten by his own name!
        with self.app.test_request_context('/user/',
                                           method='PUT',
                                           query_string=urlencode(
                                               {"user": "wordy2",
                                                "resolver": resolver,
                                                "password": "newPassword"}),
                                           headers={'Authorization': wordy_auth_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

        # Although the user "wordy" tried to update the password of user
        # "wordy2", he updated his own password.
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "wordy@{0!s}".format(realm),
                                                 "password": "newPassword"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "user", result)

        # The administrator can update the username of the user by specifying the userid
        with self.app.test_request_context('/user/',
                                           method='PUT',
                                           query_string=urlencode(
                                               {"user": "wordy2",
                                                "resolver": resolver,
                                                "userid": "7"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

        # Delete the users
        with self.app.test_request_context('/user/{0!s}/{1!s}'.format(resolver, "wordy2"),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

        with self.app.test_request_context('/user/{0!s}/{1!s}'.format(resolver, "wordy"),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

    def test_03_create_update_delete_unicode_user(self):
        realm = "sqlrealm"
        resolver = "SQL1"
        parameters = self.parameters
        parameters["resolver"] = resolver
        parameters["type"] = "sqlresolver"

        rid = save_resolver(parameters)
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(realm, [{'name': resolver}])
        self.assertEqual(len(failed), 0)
        self.assertEqual(len(added), 1)

        # CREATE a user
        with self.app.test_request_context('/user/',
                                           method='POST',
                                           data={"user": "wördy",
                                                 "resolver": resolver,
                                                 "surname": "zappa",
                                                 "givenname": "frank",
                                                 "email": "f@z.com",
                                                 "phone": "12345",
                                                 "mobile": "12345",
                                                 "password": "12345"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value") > 6, result.get("value"))

        # Get users
        with self.app.test_request_context('/user/',
                                           method='GET',
                                           query_string=urlencode(
                                               {"username": "wördy".encode('utf-8')}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"))
            self.assertEqual(result.get("value")[0].get("username"), "wördy")

        # Update by administrator. Set the password to "passwort"
        with self.app.test_request_context('/user/',
                                           method='PUT',
                                           query_string=urlencode(
                                               {"user": "wördy".encode('utf-8'),
                                                "resolver": resolver,
                                                "password": "passwort"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

        # Get user authentication and update by user.
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "wördy@{0!s}".format(realm).encode('utf-8'),
                                                 "password": "passwort"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            wordy_auth_token = result.get("value").get("token")
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "user", result)

        # Even if a user specifies another username, the username is
        # overwritten by his own name!
        with self.app.test_request_context('/user/',
                                           method='PUT',
                                           query_string=urlencode(
                                               {"user": "wördy2".encode('utf-8'),
                                                "resolver": resolver,
                                                "password": "newPassword"}),
                                           headers={'Authorization': wordy_auth_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))

        # Although the user "wördy" tried to update the password of user
        # "wördy2", he updated his own password.
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "wördy@{0!s}".format(realm).encode('utf-8'),
                                                 "password": "newPassword"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "user", result)

        # Delete the users
        with self.app.test_request_context(quote(f"/user/{resolver}/wördy"),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"))
        # Check that the user is removed from the resolver
        res = get_resolver_object(resolver)
        uid = res.getUserId("wördy")
        self.assertEqual(uid, "")

    def test_10_additional_attributes(self):
        from privacyidea.lib.policy import set_policy, ACTION, SCOPE, delete_policy
        with self.app.test_request_context('/user/attribute',
                                           method='POST',
                                           data={"user": "cornelius@realm1",
                                                 "key": "newattribute",
                                                 "value": "newvalue"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 403, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), res.data)
            self.assertEqual(result.get("error").get("message"),
                             "You are not allowed to set the custom user attribute newattribute!")

        # Allow to set custom attributes
        set_policy("custom_attr", scope=SCOPE.ADMIN,
                   action="{0!s}=:*:*".format(ACTION.SET_USER_ATTRIBUTES))

        # Check that the user has not attribute
        with self.app.test_request_context('/user/attribute',
                                           method='GET',
                                           query_string={"user": "cornelius@realm1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertNotIn("newattribute", result.get("value"))

        with self.app.test_request_context('/user/attribute',
                                           method='POST',
                                           data={"user": "cornelius@realm1",
                                                 "key": "newattribute",
                                                 "value": "newvalue"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertTrue(result.get("value") >= 0)

        # Now we verify if the user has the additional attribute:
        with self.app.test_request_context('/user/attribute',
                                           method='GET',
                                           query_string={"user": "cornelius@realm1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertIn("newattribute", result.get("value"))
            self.assertEqual(result.get("value").get("newattribute"), "newvalue")

        with self.app.test_request_context('/user/attribute',
                                           method='GET',
                                           query_string={"user": "cornelius@realm1",
                                                         "key": "newattribute"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertEqual(result.get("value"), "newvalue")

        # Now we check, if the additional attribute is also contained in the
        # user listing
        delete_policy("custom_attr")
        with self.app.test_request_context('/user/',
                                           method='GET',
                                           query_string={"realm": "realm1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            additional_attribute_found = False
            # check in the user list for the username=cornelius
            for user in result.get("value"):
                if user.get("username") == "cornelius":
                    self.assertEqual(user.get("newattribute"), "newvalue")
                    additional_attribute_found = True
            self.assertTrue(additional_attribute_found)

        # Now we search for the one explicit user
        with self.app.test_request_context('/user/',
                                           method='GET',
                                           query_string={"realm": "realm1",
                                                         "username": "cornelius"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # check in the user list for the username=cornelius
            self.assertEqual(len(result.get("value")), 1)
            self.assertEqual(result.get("value")[0].get("newattribute"), "newvalue")

        # The additional attribute should also be returned, if the user authenticates successfully.
        init_token({"serial": "SPASS1", "type": "spass", "pin": "test"},
                   user=User("cornelius", self.realm1))
        set_policy(name="POL1", scope=SCOPE.AUTHZ, action=ACTION.ADDUSERINRESPONSE)
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "cornelius@realm1",
                                                 "pass": "test"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            details = res.json.get("detail")
            user_data = details.get("user")
            self.assertIn("newattribute", user_data)
            self.assertEqual(user_data.get("newattribute"), "newvalue")
        remove_token("SPASS1")
        delete_policy("POL1")

        # Try to delete custom user attributes without a policy
        with self.app.test_request_context('/user/attribute/newattribute/cornelius/realm1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 403, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), res.data)
            error = result.get("error")
            self.assertEqual("You are not allowed to delete the custom user attribute newattribute!",
                             error.get("message"))

        # Now we delete the additional user attribute
        set_policy("custom_attr", scope=SCOPE.ADMIN,
                   action="{0!s}=*".format(ACTION.DELETE_USER_ATTRIBUTES))
        with self.app.test_request_context('/user/attribute/newattribute/cornelius/realm1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertTrue(result.get("value") >= 0)

        # and verify, that it is gone
        with self.app.test_request_context('/user/attribute',
                                           method='GET',
                                           query_string={"user": "cornelius@realm1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertNotIn("newattribute", result.get("value"))

        # The admin is allowed to delete all attributes. Check what happens
        # if the admin tries to delete an attribute that does not exist:
        # Returns a result-value: 0
        with self.app.test_request_context('/user/attribute/doesnotexist/cornelius/realm1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertTrue(result.get("value") == 0)

        # Check, which attributes the admin is allowed to set or delete
        set_policy("custom_attr", scope=SCOPE.ADMIN,
                   action="{0!s}=:hello: one two".format(ACTION.SET_USER_ATTRIBUTES))
        set_policy("custom_attr2", scope=SCOPE.ADMIN,
                   action="{0!s}=:hello2: * :hello: three".format(ACTION.SET_USER_ATTRIBUTES))
        set_policy("custom_attr3", scope=SCOPE.ADMIN,
                   action="{0!s}=:*: on off".format(ACTION.SET_USER_ATTRIBUTES))
        set_policy("custom_attr4", scope=SCOPE.ADMIN,
                   action="{0!s}=*".format(ACTION.DELETE_USER_ATTRIBUTES))
        with self.app.test_request_context('/user/editable_attributes/',
                                           method='GET',
                                           query_string={"user": "cornelius@realm1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            value = result.get("value")
            self.assertIn("delete", value)
            self.assertEqual(value.get("delete"), ['*'])
            self.assertIn("set", value)
            setables = value.get("set")
            self.assertIn("*", setables)
            self.assertIn("hello", setables)
            self.assertIn("hello2", setables)
            self.assertEqual(["off", "on"], sorted(setables.get("*")))
            self.assertIn("one", setables.get("hello"))
            self.assertIn("two", setables.get("hello"))
            self.assertIn("three", setables.get("hello"))
            self.assertEqual(["*"], setables.get("hello2"))

        set_policy("custom_create_user", scope=SCOPE.ADMIN,
                   action=ACTION.ADDUSER)
        # CREATE a user
        self._create_user_wordy()
        self._get_wordy_auth_token()
        # The admin sets attributes for this user:
        with self.app.test_request_context('/user/attribute',
                                           method='POST',
                                           data={"user": "wordy@sqlrealm",
                                                 "key": "hello",
                                                 "value": "one"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertTrue(result.get("value") >= 0)

        # We let the user wordy@sqlrealm login and GET his attributes
        with self.app.test_request_context('/user/attribute',
                                           method='GET',
                                           query_string={"user": "cornelius@realm1"},
                                           headers={'Authorization': self.wordy_auth_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.assertIn("hello", result.get("value"))
            self.assertEqual("one", result.get("value").get("hello"))

        # The user tries to delete his attribute, but he is not allowed to.
        with self.app.test_request_context('/user/attribute/hello/wordy/sqlrealm',
                                           method='DELETE',
                                           headers={'Authorization': self.wordy_auth_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 403, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), res.data)
            error = result.get("error")
            self.assertEqual("You are not allowed to delete the custom user attribute hello!",
                             error.get("message"))

        # The user tries to set an attribute, but he is not allowed to.
        with self.app.test_request_context('/user/attribute',
                                           method='POST',
                                           data={"key": "newattr",
                                                 "value": "newvalue"},
                                           headers={'Authorization': self.wordy_auth_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 403, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), res.data)
            error = result.get("error")
            self.assertEqual("You are not allowed to set the custom user attribute newattr!",
                             error.get("message"))

        delete_policy("custom_attr")
        delete_policy("custom_attr2")
        delete_policy("custom_attr3")
        delete_policy("custom_attr4")
        delete_policy("custom_create_user")



#  2018-26-09 Paul Lettich <paul.lettich@netknights.it>
#             Add decrease/reset functions
#  2018-03-01 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
#  Copyright (C) 2018 Cornelius Kölbel
#  License:  AGPLv3
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module is used to modify counters in the database
"""
from sqlalchemy import func

from privacyidea.lib.config import get_privacyidea_node
from privacyidea.models import EventCounter, db


def increase(counter_name):
    """
    Increase the counter value in the database.
    If the counter does not exist yet, create the counter.

    :param counter_name: The name/identifier of the counter
    :return: None
    """
    # If there is no table row for the current node, create one.
    node = get_privacyidea_node()
    counter = EventCounter.query.filter_by(counter_name=counter_name, node=node).first()
    if not counter:
        counter = EventCounter(counter_name, 0, node=node)
        counter.save()
    counter.increase()


def _reset_counter_on_all_nodes(counter_name):
    """
    Reset all EventCounter rows that set a value for ``counter_name`` to zero,
    regardless of the node column.
    :param counter_name:  The name/identifier of the counter
    """
    EventCounter.query.filter_by(counter_name=counter_name).update({'counter_value': 0})
    db.session.commit()


def decrease(counter_name, allow_negative=False):
    """
    Decrease the counter value in the database.
    If the counter does not exist yet, create the counter.
    Also checks whether the counter is allowed to become negative.

    :param counter_name: The name/identifier of the counter
    :param allow_negative: Whether the counter can become negative. Note that even if this flag is not set,
                           the counter may become negative due to concurrent queries.
    :return: None
    """
    node = get_privacyidea_node()
    counter = EventCounter.query.filter_by(counter_name=counter_name, node=node).first()
    if not counter:
        counter = EventCounter(counter_name, 0, node=node)
        counter.save()
    # We are allowed to decrease the current counter object only if the overall
    # counter value is positive (because individual rows may be negative then),
    # or if we allow negative values. Otherwise, we need to reset all rows of all nodes.
    if read(counter_name) > 0 or allow_negative:
        counter.decrease()
    else:
        _reset_counter_on_all_nodes(counter_name)


def reset(counter_name):
    """
    Reset the counter value in the database to 0.
    If the counter does not exist yet, create the counter.
    :param counter_name: The name/identifier of the counter
    :return:
    """
    node = get_privacyidea_node()
    counters = EventCounter.query.filter_by(counter_name=counter_name).count()
    if not counters:
        counter = EventCounter(counter_name, 0, node=node)
        counter.save()
    else:
        _reset_counter_on_all_nodes(counter_name)


def read(counter_name):
    """
    Read the counter value from the database.
    If the counter_name does not exist, 'None' is returned.

    :param counter_name: The name of the counter
    :return: The value of the counter
    """
    return db.session.query(func.sum(EventCounter.counter_value))\
        .filter(EventCounter.counter_name == counter_name).one()[0]



"""Add a generic admin policy for all admin read actions

Revision ID: 3d7f8b29cbb1
Revises: dceb6cd3c41e
Create Date: 2019-07-04 22:37:53.594962

"""

# revision identifiers, used by Alembic.
revision = '3d7f8b29cbb1'
down_revision = 'dceb6cd3c41e'

from alembic import op
import sqlalchemy as sa

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import orm
from sqlalchemy.schema import Sequence
from privacyidea.lib.policy import SCOPE, ACTION

Base = declarative_base()

# dashes are not allowed, when creating policies in the WebUI
# or via the library. So we are sure, that in normal operation
# this policy can never be created.
POLICYNAME = u"pi-update-policy-3d7f8b29cbb1"


class Policy(Base):
    """
    The policy table contains policy definitions which control
    the behaviour during
     * enrollment
     * authentication
     * authorization
     * administration
     * user actions
    """
    __tablename__ = "policy"
    __table_args__ = {'mysql_row_format': 'DYNAMIC'}
    id = sa.Column(sa.Integer, Sequence("policy_seq"), primary_key=True)
    active = sa.Column(sa.Boolean, default=True)
    check_all_resolvers = sa.Column(sa.Boolean, default=False)
    name = sa.Column(sa.Unicode(64), unique=True, nullable=False)
    scope = sa.Column(sa.Unicode(32), nullable=False)
    action = sa.Column(sa.Unicode(2000), default=u"")
    realm = sa.Column(sa.Unicode(256), default=u"")
    adminrealm = sa.Column(sa.Unicode(256), default=u"")
    resolver = sa.Column(sa.Unicode(256), default=u"")
    user = sa.Column(sa.Unicode(256), default=u"")
    client = sa.Column(sa.Unicode(256), default=u"")
    time = sa.Column(sa.Unicode(64), default=u"")
    # If there are multiple matching policies, choose the one
    # with the lowest priority number. We choose 1 to be the default priotity.
    priority = sa.Column(sa.Integer, default=1, nullable=False)


def upgrade():
    """
    During upgrade we check, if admin policies exist.
    If so, we add a generic policy for all admins, that allows to
    read all configuration, which mimicks the previous behaviour
    :return:
    """
    actions = u",".join([ACTION.PERIODICTASKREAD, ACTION.MACHINERESOLVERREAD,
                         ACTION.PRIVACYIDEASERVERREAD, ACTION.RADIUSSERVERREAD,
                         ACTION.SMTPSERVERREAD, ACTION.EVENTHANDLINGREAD,
                         ACTION.POLICYREAD, ACTION.RESOLVERREAD,
                         ACTION.CACONNECTORREAD, ACTION.SMSGATEWAYREAD,
                         ACTION.STATISTICSREAD, ACTION.SYSTEMREAD])

    bind = op.get_bind()
    session = orm.Session(bind=bind)
    if session.query(Policy.id).filter(Policy.scope == u"{0!s}".format(SCOPE.ADMIN),
                                       Policy.active.is_(True)).all():

        if session.query(Policy.id).filter_by(name=POLICYNAME).first() is None:
            # add policy
            tokenlist_pol = Policy(name=POLICYNAME, scope=u"{0!s}".format(SCOPE.ADMIN),
                                   action=actions)
            session.add(tokenlist_pol)
            print("Added '{0!s}' action for admin policies.".format(actions))
        else:
            print("Policy {} already exists.".format(POLICYNAME))
    else:
        print("No admin policy active. No need to create '{0!s}' action.".format(actions))

    try:
        session.commit()
    except Exception as exx:
        print("Could not create policy {}: {!r}".format(POLICYNAME, exx))
        print(exx)


def downgrade():
    # Delete the policy, if it still exists
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    session.query(Policy).filter(Policy.name == u"{0!s}".format(POLICYNAME)).delete()
    session.commit()



# SPDX-FileCopyrightText: (C) 2024 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# 2024-11-11 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#            New pi-token-janitor script
# 2023-11-03 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#            Migrate to click
# 2020-11-11 Timo Sturm <timo.sturm@netknights.it>
#            Select how to validate PSKC imports
# 2018-02-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Allow to import PSKC file
# 2017-11-21 Cornelius Kölbel <corenlius.koelbel@netknights.it>
#            export to CSV including usernames
# 2017-10-18 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add token export (HOTP and TOTP) to PSKC
# 2017-05-02 Friedrich Weber <friedrich.weber@netknights.it>
#            Improve token matching
# 2017-04-25 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys

from datetime import datetime
from dateutil import parser
import re
from flask.cli import AppGroup

import click
from collections import defaultdict
from dateutil.tz import tzlocal
from typing import Generator, Callable, Union
from yaml import safe_dump as yaml_safe_dump

from privacyidea.lib.container import find_container_for_token
from privacyidea.lib.error import ResolverError
from privacyidea.lib.importotp import export_pskc
from privacyidea.lib.utils import parse_legacy_time, is_true

from privacyidea.models import Token, TokenContainer
from privacyidea.lib.token import unassign_token, remove_token, get_tokens_paginated_generator
from privacyidea.lib.tokenclass import TokenClass

allowed_tokenattributes = [col.key for col in Token.__table__.columns]

comparator_pattern = re.compile(r"^\s*([^!=<>]+?)\s*([!=<>])\s*([^!=<>]+?)\s*$")


def _try_convert_to_integer(given_value_string: str) -> int:
    try:
        return int(given_value_string)
    except ValueError:
        raise click.ClickException(f'Not an integer: {given_value_string}')


def _try_convert_to_datetime(given_value_string: str) -> datetime:
    try:
        parsed = parser.parse(given_value_string, dayfirst=False)
        if not parsed.tzinfo:
            # If not timezone is given we assume the timestamp is given in local time
            parsed = parsed.replace(tzinfo=tzlocal())
        return parsed
    except ValueError:
        raise


def _compare_regex_or_equal(given_regex: str) -> Callable[[Union[int, bool, str]], bool]:
    def comparator(value: Union[int, bool, str]) -> bool:
        if type(value) in (int, bool):
            # If the value from the database is an integer, we compare "equals integer"
            given_value = _try_convert_to_integer(given_regex)
            return given_value == value
        else:
            # if the value from the database is a string, we compare regex
            return bool(re.search(given_regex, value))

    return comparator


def _compare_not(given_regex: str) -> Callable[[Union[int, bool, str]], bool]:
    def comparator(value: Union[int, bool, str]) -> bool:
        if type(value) in (int, bool):
            # If the value from the database is an integer, we compare "equals integer"
            given_value = _try_convert_to_integer(given_regex)
            return given_value != value
        else:
            # if the value from the database is a string, we compare regex
            return not re.search(given_regex, value)

    return comparator


def _compare_greater_than(given_value: Union[int, str]) -> Callable[[int], bool]:
    """
    :return: a function which returns True if its parameter (converted to an integer)
             is greater than *given_value_string* (converted to an integer).
    """
    given_value = int(given_value)

    def comparator(value: int) -> bool:
        try:
            return int(value) > given_value
        except ValueError:
            return False

    return comparator


def _compare_less_than(given_value: Union[int, str]) -> Callable[[int], bool]:
    """
    :return: a function which returns True if its parameter (converted to an integer)
             is less than *given_value_string* (converted to an integer).
    """
    given_value = int(given_value)

    def comparator(value: int) -> bool:
        try:
            return int(value) < given_value
        except ValueError:
            return False

    return comparator


def _compare_after(given_value: datetime) -> Callable[[str], bool]:
    """
    :return: a function which returns True if its parameter (converted to a datetime) occurs after
             given_value.
    """
    def comparator(value: str):
        try:
            return parse_legacy_time(value, return_date=True) > given_value
        except ValueError:
            return False
    return comparator


def _compare_before(given_value: datetime) -> Callable[[str], bool]:
    """
    :return: a function which returns True if its parameter (converted to a datetime) occurs before
             given_value.
    """
    def comparator(value: str):
        try:
            return parse_legacy_time(value, return_date=True) < given_value
        except ValueError:
            return False

    return comparator


def build_filter(filter_string: str, allowed_keys: list = None) -> tuple:
    """
    Build and return a filter closure, which is based on the given filter_string.
    The filter closure takes a value and returns True if the user-defined criterion matches.

    :param: filter_string: a filter string which contains the key, the operator and the value
    :type filter_string: str
    :param allowed_keys: list of allowed values for keys
    :type allowed_keys: list
    :return: a tuple with the given key and comparator closure with the given value
    :rtype: func
    """
    match = comparator_pattern.match(filter_string)
    if not match:
        raise click.ClickException(f'The option requires a filter '
                                   f'syntax like: "<key> = <value>" with '
                                   f'possible comparators "=", "<", ">" and "!". '
                                   f'(given: "{filter_string}")')
    if allowed_keys and match.group(1) not in allowed_keys:
        raise click.ClickException(f"Key '{match.group(1)}' not allowed. "
                                   f"Allowed values for keys are '{allowed_keys}'")
    if match.group(2) == '=':
        return match.group(1), _compare_regex_or_equal(match.group(3))
    elif match.group(2) == '<':
        try:  # first try to convert the given value to an integer
            given_value = int(match.group(3))
            return match.group(1), _compare_less_than(given_value)
        except ValueError:
            try:  # then we try to parse as a datetime object
                given_value = _try_convert_to_datetime(match.group(3))
                return match.group(1), _compare_before(given_value)
            except ValueError:
                raise click.ClickException(f"Unable to find a comparator for {filter_string}")
    elif match.group(2) == '>':
        try:  # first try to convert the given value to an integer
            given_value = int(match.group(3))
            return match.group(1), _compare_greater_than(given_value)
        except ValueError:
            try:  # try to convert the given value to datetime
                given_value = _try_convert_to_datetime(match.group(3))
                return match.group(1), _compare_after(given_value)
            except ValueError:
                raise click.ClickException(f"Unable to find a comparator for {filter_string}")
    elif match.group(2) == '!':
        return match.group(1), _compare_not(match.group(3))
    else:
        raise click.ClickException('Invalid matching operator. Use "=", ">", "<" or "!"')


filter_funcs = {"=": _compare_regex_or_equal,
                "<": _compare_less_than,
                ">": _compare_greater_than,
                "!": _compare_not}


def build_token_attribute_filter(tokenattribute: str) -> tuple:
    """
    Build and return a token attribute filter.
    The tokenattribute is separated into its components and the appropriate
    filter closure is returned with the parsed value.

    :param: tokenattribute: a string which contains the key, the operator and the value
    :type tokenattribute: str
    :return: tuple of the parsed attribute and a comparator function
    :rtype: tuple
    """
    match = comparator_pattern.match(tokenattribute)
    if not match:
        raise click.ClickException(f'The tokenattribute option requires a filter '
                                   f'syntax like: "<attribute> = <value>" with '
                                   f'possible comparators "=", "<", ">" and "!". '
                                   f'(given: "{tokenattribute}")')
    if match.group(1) not in allowed_tokenattributes:
        raise click.ClickException(f"Token attribute {match.group(1)} not allowed! "
                                   f"Allowed token attributes are: {allowed_tokenattributes}")
    else:
        return match.group(1), filter_funcs[match.group(2)](match.group(3))


def export_token_data(token_list: list, token_attributes: list = None,
                      user_attributes: list = None) -> list:
    """
    Returns a list of tokens. Each token again is a dictionary of the requested
    token attributes, tokeninfo and user attributes

    :param token_list: List of token objects to export
    :param token_attributes: add additional token attributes
    :param user_attributes: add additional user attributes
    :return: a list of token dictionaries with the requested attributes
    :rtype: list
    """
    tokens = []
    for token_obj in token_list:
        token_data = {"serial": f'{token_obj.token.serial}',
                      "tokentype": f'{token_obj.token.tokentype}',
                      "realms": f'{token_obj.get_realms()}'}
        token_info = token_obj.get_tokeninfo()
        export_ti = {}
        if token_attributes:
            for att in token_attributes:
                if att in allowed_tokenattributes:
                    token_data[att] = f'{token_obj.token.get(att)}'
                else:
                    if att in token_info:
                        export_ti[att] = f'{token_info[att]}'
            token_data["info"] = export_ti
        else:
            # Return all token attributes/tokeninfo
            token_data["info"] = token_info

        if user_attributes:
            try:
                user_data = {}
                user = token_obj.user
                if user:
                    for att in user_attributes:
                        if att == "uid":
                            user_data[att] = f'{user.uid}'
                        elif att == "resolver":
                            user_data[att] = f'{user.resolver}'
                        elif att == "realm":
                            user_data[att] = f'{user.realm}'
                        else:
                            if att in user.info:
                                user_data[att] = f'{user.info[att]}'
                token_data["user"] = user_data
            except Exception as e:
                token_data["user"] = {"error": f"failed to resolve user: {e}"}
        tokens.append(token_data)
    return tokens


def export_user_data(token_list: list, user_attributes: list = None) -> dict:
    """
    Returns a list of users with the information how many tokens this user has assigned

    :param token_list:
    :param user_attributes: display additional user attributes
    :return:
    :rtype: dict
    """
    users = {}
    for token_obj in token_list:
        try:
            user = token_obj.user
        except Exception as e:
            sys.stderr.write(f"Failed to determine user for token {token_obj.token.serial} ({e}.\n")
            user = None
        if user:
            uid = (f"'{user.info.get('username', '')}','{user.info.get('givenname', '')}',"
                   f"'{user.info.get('surname', '')}','{user.uid}','{user.resolver}','{user.realm}'")
            if user_attributes:
                for att in user_attributes:
                    uid += f",'{user.info.get(att, '')}'"
        else:
            uid = "N/A" + ", " * 5

        if uid in users.keys():
            users[uid].append(token_obj.token.serial)
        else:
            users[uid] = [token_obj.token.serial]

    return users


def _get_tokenlist(assigned: Union[bool, None], active: Union[bool, None], range_of_serial: str,
                   tokeninfo_filter, tokenattribute_filter: list,
                   tokenowner_filter, tokencontaner_filter, tokentype, realm, resolver, rollout_state,
                   orphaned: Union[bool, None], chunksize: int, has_not_tokeninfo_key, has_tokeninfo_key,
                   orphaned_on_error: bool = False) -> Generator[TokenClass, None, None]:
    if assigned is not None:
        assigned = is_true(assigned)
    if active is not None:
        active = is_true(active)

    iterable = get_tokens_paginated_generator(tokentype=tokentype,
                                              realm=realm,
                                              resolver=resolver,
                                              rollout_state=rollout_state,
                                              active=active,
                                              assigned=assigned,
                                              psize=chunksize)

    for token_object_list in iterable:
        filtered_list = []
        for token_obj in token_object_list:
            add = True
            # TODO: We could do this with regex and the tokeninfo filter
            if has_not_tokeninfo_key:
                if has_not_tokeninfo_key in token_obj.get_tokeninfo():
                    add = False
            if has_tokeninfo_key:
                if has_tokeninfo_key not in token_obj.get_tokeninfo():
                    add = False
            if range_of_serial:
                if not range_of_serial[0] <= token_obj.token.serial <= range_of_serial[1]:
                    add = False
            if tokeninfo_filter:
                for tokeninfo_key, tokeninfo_comparator in tokeninfo_filter:
                    value = token_obj.get_tokeninfo(tokeninfo_key)
                    # if the tokeninfo key is not even set, it does not match the filter
                    if value is None:
                        add = False
                    elif not tokeninfo_comparator(value):
                        add = False
            if tokenattribute_filter:
                for tokenattribute_key, tokenattribute_comparator in tokenattribute_filter:
                    value = token_obj.token.get(tokenattribute_key)
                    if value is None:
                        add = False
                    elif not tokenattribute_comparator(value):
                        add = False
            if tokenowner_filter:
                try:
                    user = token_obj.user
                    if user is not None:
                        # First check for attributes of the user object
                        for tokenowner_key, tokenowner_comparator in tokenowner_filter:
                            if tokenowner_key in ["uid", "resolver", "realm", "login"]:
                                value = getattr(user, tokenowner_key)
                            else:
                                value = user.info.get(tokenowner_key)
                            if value is None:
                                add = False
                            elif not tokenowner_comparator(value):
                                add = False
                    else:
                        add = False
                except ResolverError:
                    # Unable to resolve user, no filter applicable
                    add = False
            if tokencontaner_filter:
                for tokenowner_key in tokencontaner_filter:
                    container = find_container_for_token(token_obj.token.serial)
                    if container is not None:
                        container_info = vars(container)
                        container_info["serial"] = container.serial
                        container_info["type"] = container.type
                        container_info["description"] = container.description
                        container_info["realm"] = container.realms
                        container_info["last_seen"] = container.last_seen
                        container_info["last_update"] = container.last_updated
                        value = container_info.get(tokenowner_key[0])
                        if value is None:
                            add = False
                        elif not all(comparator(value) for comparator in tokenowner_key[1]):
                            add = False
                    else:
                        add = False
            if orphaned is not None:
                if is_true(orphaned):
                    try:
                        if token_obj.user is None or token_obj.user.exist():
                            # Either the token has no user assigned or the
                            # assigned user exists in the resolver
                            add = False
                    except ResolverError:
                        # Could not resolve user due to errors in the resolver.
                        # Don't mark the token as orphaned.
                        add = orphaned_on_error
                else:
                    try:
                        if token_obj.user is not None and not token_obj.user.exist():
                            add = False
                    except ResolverError:
                        add = not orphaned_on_error

            if add:
                # if everything matched, we append the token object
                filtered_list.append(token_obj)

        yield filtered_list


@click.group('find', invoke_without_command=True, cls=AppGroup)
@click.option('--chunksize', default=1000, show_default=True,
              help='The number of tokens to fetch in one request.')
# TODO: Maby remove has-not-tokeninfo-key and has-tokeninfo-key and use regex instead
@click.option('--has-not-tokeninfo-key',
              help='filters for tokens that have not given the specified tokeninfo-key')
@click.option('--has-tokeninfo-key',
              help='filters for tokens that have given the specified tokeninfo-key.')
@click.option('--tokeninfo', 'tokeninfos', multiple=True,
              help='Match for a certain tokeninfo from the database.')
@click.option('--tokenattribute', 'tokenattributes', multiple=True,
              help='Match for a certain token attribute from the database. You can use the following operators: '
                   '=, >, <, ! for you matching. For example: "rollout_state=clientwait" or "failcount>3".')
@click.option('--tokenowner', 'tokenowners', multiple=True,
              help='Match for certain information of the token owner from the database. '
                   'Example: user_id=642cf598-d9cf-1037-8083-a1df7d38c897.')
@click.option('--tokencontainer', 'tokencontainers', multiple=True,
              help='Match for certain information of tokencontainer from the database. Example: type=smartphone.')
@click.option('--assigned',
              help='Whether the token is assigned to a user. Can be "True" or "False"')
@click.option('--active',
              help='Whether to token is active/enabled. Can be "True" or "False"')
@click.option('--orphaned',
              help='Whether the token is an orphaned token. Can be "True" or "False"')
@click.option('--orphaned-on-error', is_flag=True, default=False,
              help="Mark token as orphaned if an error occurred when resolving user.")
@click.option('--range-of-serial', help='A range of serial numbers to search for. '
                                        'For Example HOTP10000000-HOTP20000000')
@click.pass_context
def findtokens(ctx, chunksize, has_not_tokeninfo_key, has_tokeninfo_key,
               tokenattributes, tokeninfos, tokenowners, tokencontainers,
               assigned, active, orphaned, range_of_serial, orphaned_on_error):
    """
    Find all tokens which match the given conditions.
    """
    if range_of_serial:
        range_of_serial = range_of_serial.split("-")
    ta_filter = []
    if tokenattributes:
        for tokenattribute in tokenattributes:
            ta_filter.append(build_token_attribute_filter(tokenattribute))

    ti_filter = []
    if tokeninfos:
        for tokeninfo in tokeninfos:
            ti_filter.append(build_filter(tokeninfo))

    to_filter = []
    if tokenowners:
        for owner in tokenowners:
            to_filter.append(build_filter(owner))

    tc_filter = []
    if tokencontainers:
        allowed_container_keys = [col.key for col in TokenContainer.__table__.columns]
        for container in tokencontainers:
            tc_filter.append(build_filter(container, allowed_container_keys))

    ctx.obj = dict()

    ctx.obj['tokens'] = _get_tokenlist(assigned=assigned, active=active, range_of_serial=range_of_serial,
                                       tokeninfo_filter=ti_filter, tokenattribute_filter=ta_filter,
                                       tokenowner_filter=to_filter, tokencontaner_filter=tc_filter,
                                       tokentype=None, realm=None, resolver=None,
                                       rollout_state=None, orphaned=orphaned,
                                       chunksize=chunksize, has_not_tokeninfo_key=has_not_tokeninfo_key,
                                       has_tokeninfo_key=has_tokeninfo_key,
                                       orphaned_on_error=orphaned_on_error)

    if ctx.invoked_subcommand is None:
        ctx.invoke(list_cmd)


@findtokens.command('list')
@click.option('-u', '--show-user-attribute', 'user_attributes', multiple=True,
              help='Show additional user attributes (can be given multiple times).')
@click.option('-t', '--show-token-attribute', 'token_attributes', multiple=True,
              help='Show additional token attributes/tokeninfo values (can be given '
                   'multiple times). The default is to show all tokeninfo values.')
@click.option('-s', '--summarize', 'sum_tokens', is_flag=True, default=False,
              help='Reduce the output to show only the number of tokens owned by each user.')
@click.pass_context
def list_cmd(ctx, user_attributes, token_attributes, sum_tokens):
    """
    List all found tokens.
    """
    users_sum = defaultdict(int)
    for tlist in ctx.obj['tokens']:
        if not sum_tokens:
            tokens = export_token_data(tlist, token_attributes=token_attributes,
                                       user_attributes=user_attributes)
            for token in tokens:
                click.echo(token)
        else:
            users = export_user_data(tlist, user_attributes)
            for user, tokens in users.items():
                users_sum[user] += len(tokens)
    if sum_tokens:
        for user, count in users_sum.items():
            click.echo(f"{user},{count}")


@findtokens.command('export')
@click.option('--format', 'export_format',
              type=click.Choice(['csv', 'yaml', 'pskc'], case_sensitive=False),
              default='pskc', show_default=True,
              help='The output format of the token export. CSV export only '
                   'allows TOTP and HOTP token types')
@click.option('--b32', is_flag=True,
              help='In case of exporting tokens to CSV or YAML, the seed is '
                   'written as base32 encoded instead of hex.')
# TODO: check if there is a better way
@click.pass_context
def export(ctx, export_format, b32):
    """
    Export found tokens.
    """
    for tlist in ctx.obj['tokens']:
        if export_format == "csv":
            for tokenobj in tlist:
                if tokenobj.type.lower() not in ["totp", "hotp"]:
                    continue
                token_dict = tokenobj._to_dict(b32=b32)
                owner = f"{tokenobj.user.login}@{tokenobj.user.realm}" if tokenobj.user else "n/a"
                export_string = (f"{owner}, {token_dict.get('serial')}, {token_dict.get('otpkey')}, "
                                 f"{token_dict.get('type')}, {token_dict.get('otplen')}")
                if tokenobj.type.lower() == "totp":
                    click.echo(export_string + f", {token_dict.get('info_list', {}).get('timeStep')}")
                else:
                    click.echo(export_string)
        elif export_format == "yaml":
            token_list = []
            for tokenobj in tlist:
                try:
                    token_dict = tokenobj._to_dict(b32=b32)
                    token_dict["owner"] = f"{tokenobj.user.login}@{tokenobj.user.realm}" if tokenobj.user else "n/a"
                    token_list.append(token_dict)
                except Exception as e:
                    sys.stderr.write(f"\nFailed to export token {tokenobj.get_serial()} ({e}).\n")
            click.echo(yaml_safe_dump(token_list))
        else:
            key, token_num, soup = export_pskc(tlist)
            sys.stderr.write(f"\n{token_num} tokens exported.\n")
            sys.stderr.write(f"\nThis is the AES encryption key of the token seeds.\n"
                             "You need this key to import the "
                             "tokens again:\n\n\t{key}\n\n")
            click.echo(f"{soup}")


@findtokens.command('set_tokenrealms')
@click.option('--tokenrealm', 'tokenrealms', multiple=True, required=True, type=str)
@click.pass_context
def set_tokenrealms(ctx, tokenrealms):
    """
    Set the realms of the found tokens.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            trealms = [r.strip() for r in tokenrealms if r]
            token_obj.set_realms(trealms)
            click.echo(f"Setting realms of token {token_obj.token.serial} to {trealms}.")


@findtokens.command('disable')
@click.pass_context
def disable(ctx):
    """
    Disable found tokens.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            token_obj.enable(enable=False)
            token_obj.save()
            click.echo(f"Disabled token {token_obj.token.serial}")


@findtokens.command('enable')
@click.pass_context
def enable(ctx):
    """
    Enable found tokens.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            token_obj.enable(enable=True)
            token_obj.save()
            click.echo(f"Enabled token {token_obj.token.serial}")


@findtokens.command('delete')
@click.pass_context
def delete(ctx):
    """
    Delete found tokens.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            remove_token(serial=token_obj.token.serial)
            click.echo(f"Deleted token {token_obj.token.serial}")


@findtokens.command('unassign')
@click.pass_context
def unassign(ctx):
    """
    Unassign the found tokens from their owners.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            unassign_token(serial=token_obj.token.serial)
            click.echo(f"Unassigned token {token_obj.token.serial}")


@findtokens.command('set_description')
@click.option('--description', required=True, type=str)
@click.pass_context
def set_description(ctx, description):
    """
    Sets the description of the found tokens.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            token_obj.set_description(description)
            token_obj.save()
            click.echo(f"Set description for token {token_obj.token.serial}: {description}")


@findtokens.command('set_tokeninfo')
@click.option('--tokeninfo', required=True, type=str)
@click.pass_context
def set_tokeninfo(ctx, tokeninfo):
    """
    Sets the tokeninfo of the found tokens.
    """
    match = re.match(r"\s*(\w+)\s*(=)\s*(\w+)\s*$", tokeninfo)
    if not match:
        raise click.ClickException(f"Can not parse tokeninfo to set. It should "
                                   f"be given as \"<key> = <value>\". (actual: {tokeninfo})")
    tokeninfo_key = match.group(1)
    tokeninfo_value = match.group(3)
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            token_obj.add_tokeninfo(tokeninfo_key, tokeninfo_value)
            token_obj.save()
            click.echo(f"Set tokeninfo for token {token_obj.token.serial}: {tokeninfo_key}={tokeninfo_value}")


@findtokens.command('remove_tokeninfo')
@click.option('--tokeninfo_key', required=True, type=str)
@click.pass_context
def remove_tokeninfo(ctx, tokeninfo_key):
    """
    Remove the tokeninfo of the found tokens.
    """
    for tlist in ctx.obj['tokens']:
        for token_obj in tlist:
            token_obj.del_tokeninfo(tokeninfo_key)
            token_obj.save()
            click.echo(f"Removed tokeninfo '{tokeninfo_key}' for token {token_obj.token.serial}")



#  Copyright (C) 2014 Cornelius Kölbel
#  contact:  corny@cornelinux.de
#
#  2024-06-25 Raphael Topel <raphael.topel@esh.essen.de>
#             Change AUTHTYPE.SASL_KERBEROS behaviour if upn is present in userinfo values for multidomain support
#  2018-12-14 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add censored password functionality
#  2017-12-22 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add configurable multi-value-attributes
#             with the help of Nomen Nescio
#  2017-07-20 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Fix unicode usernames
#  2017-01-23 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add certificate verification
#  2017-01-07 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Use get_info=ldap3.NONE for binds to avoid querying of subschema
#             Remove LDAPFILTER and self.reversefilter
#  2016-07-14 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Adding getUserId cache.
#  2016-04-13 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add object_classes and dn_composition to configuration
#             to allow flexible user_add
#  2016-04-10 Martin Wheldon <martin.wheldon@greenhills-it.co.uk>
#             Allow user accounts held in LDAP to be edited, providing
#             that the account they are using has permission to edit
#             those attributes in the LDAP directory
#  2016-02-22 Salvo Rapisarda
#             Allow objectGUID to be a users attribute
#  2016-02-19 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Allow objectGUID to be the uid.
#  2015-10-05 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Remove reverse_map, so that one LDAP field can map
#             to several privacyIDEA fields.
#  2015-04-16 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add redundancy with LDAP3 Server pools. Round Robin Strategy
#  2015-04-15 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Increase test coverage
#  2014-12-25 Cornelius Kölbel <cornelius@privacyidea.org>
#             Rewrite for flask migration
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This is the resolver to find users in LDAP directories like
OpenLDAP and Active Directory.

The file is tested in tests/test_lib_resolver.py
"""

import logging

import yaml
import threading
import functools

from .UserIdResolver import UserIdResolver

import ldap3
from ldap3 import MODIFY_REPLACE, MODIFY_ADD, MODIFY_DELETE
from ldap3 import Tls
from ldap3.core.exceptions import LDAPOperationResult
from ldap3.core.results import RESULT_SIZE_LIMIT_EXCEEDED
import ssl

import os.path

import traceback
from passlib.hash import ldap_salted_sha1
import hashlib
import binascii
from privacyidea.lib.framework import get_app_local_store, get_app_config_value
import datetime

from privacyidea.lib import _
from privacyidea.lib.log import log_with
from privacyidea.lib.utils import (is_true, to_bytes, to_unicode,
                                   convert_column_to_unicode)
from privacyidea.lib.error import privacyIDEAError, ResolverError
import uuid
from ldap3.utils.conv import escape_bytes
from operator import itemgetter

log = logging.getLogger(__name__)

try:
    import gssapi
    have_gssapi = True
except ImportError:
    log.info('Could not import gssapi package. Kerberos authentication not available')
    have_gssapi = False

CACHE = {}

ENCODING = "utf-8"
# The number of rounds the resolver tries to reach a responding server in the
#  pool
SERVERPOOL_ROUNDS = 2
# The number of seconds a non-responding server is removed from the server pool
SERVERPOOL_SKIP = 30
# The pooling strategy for the ldap servers
LDAP_STRATEGY = {"ROUND_ROBIN": ldap3.ROUND_ROBIN, "FIRST": ldap3.FIRST, "RANDOM": ldap3.RANDOM}
SERVERPOOL_STRATEGY = "ROUND_ROBIN"

# 1 sec == 10^9 nano secs == 10^7 * (100 nano secs)
MS_AD_MULTIPLYER = 10 ** 7
MS_AD_START = datetime.datetime(1601, 1, 1)

if os.path.isfile("/etc/privacyidea/ldap-ca.crt"):
    DEFAULT_CA_FILE = "/etc/privacyidea/ldap-ca.crt"
elif os.path.isfile("/etc/ssl/certs/ca-certificates.crt"):
    DEFAULT_CA_FILE = "/etc/ssl/certs/ca-certificates.crt"
elif os.path.isfile("/etc/ssl/certs/ca-bundle.crt"):
    DEFAULT_CA_FILE = "/etc/ssl/certs/ca-bundle.crt"
else:
    DEFAULT_CA_FILE = "/etc/privacyidea/ldap-ca.crt"

TLS_NEGOTIATE_PROTOCOL = ssl.PROTOCOL_TLS

DEFAULT_TLS_PROTOCOL = TLS_NEGOTIATE_PROTOCOL

TLS_OPTIONS_1_3 = (ssl.OP_NO_TLSv1_2, ssl.OP_NO_TLSv1_1, ssl.OP_NO_TLSv1, ssl.OP_NO_SSLv3)


class LockingServerPool(ldap3.ServerPool):
    """
    A ``ServerPool`` subclass that uses a RLock to synchronize invocations of
    ``initialize``, ``get_server`` and ``get_current_server``.

    We synchronize invocations to rule out race conditions when multiple threads
    try to manipulate the server pool state concurrently.

    We use a ``RLock`` instead of a simple ``Lock`` to avoid locking ourselves.
    """
    def __init__(self, *args, **kwargs):
        ldap3.ServerPool.__init__(self, *args, **kwargs)
        self._lock = threading.RLock()

    def initialize(self, connection):
        with self._lock:
            return ldap3.ServerPool.initialize(self, connection)

    def get_server(self, connection):
        with self._lock:
            return ldap3.ServerPool.get_server(self, connection)

    def get_current_server(self, connection):
        with self._lock:
            return ldap3.ServerPool.get_current_server(self, connection)


def get_ad_timestamp_now():
    """
    returns the current UTC time as it is used in Active Directory in the
    attribute accountExpires.
    This is 100-nano-secs since 1.1.1601

    :return: time
    :rtype: int
    """
    utc_now = datetime.datetime.utcnow()
    elapsed_time = utc_now - MS_AD_START
    total_seconds = elapsed_time.total_seconds()
    # convert this to (100 nanoseconds)
    return int(MS_AD_MULTIPLYER * total_seconds)


def trim_objectGUID(user_id):
    user_id = uuid.UUID("{{{0!s}}}".format(user_id)).bytes_le
    user_id = escape_bytes(user_id)
    return user_id


def get_info_configuration(noschemas):
    """
    Given the value of the NOSCHEMAS config option, return the value that should
    be passed as ldap3's `get_info` argument.
    :param noschemas: a boolean
    :return: one of ldap3.SCHEMA or ldap3.NONE
    """
    get_schema_info = ldap3.SCHEMA
    if noschemas:
        get_schema_info = ldap3.NONE
    log.debug("Get LDAP schema info: {0!r}".format(get_schema_info))
    return get_schema_info


def ignore_sizelimit_exception(conn, generator):
    """
    Wrapper for ``paged_search``, which (since ldap3 2.3) throws an exception if the size limit has been
    reached. This function wraps the generator and ignores this exception.

    Additionally, this checks ``conn.response`` for any leftover entries that were not yet returned
    by the generator and yields them.
    """
    last_entry = None
    while True:
        try:
            last_entry = next(generator)
            yield last_entry
        except StopIteration:
            # If the generator is exceeded, we stop
            break
        except LDAPOperationResult as e:
            # If the size limit has been reached, we stop. All other exceptions are re-raised.
            if e.result == RESULT_SIZE_LIMIT_EXCEEDED:
                # Workaround: In ldap3 <= 2.4.1, the generator may "forget" to yield some entries that
                # were transmitted just before the "size limit exceeded" message. In other words,
                # the exception is raised *before* the generator has yielded those entries.
                # These leftover entries can still be found in ``conn.response``, so we
                # just yield them here.
                # However, as future versions of ldap3 may fix this behavior and
                # may actually yield those elements as well, this workaround may result in
                # duplicate entries.
                # Thus, we check if the last entry we got from the generator can be found
                # in ``conn.response``. If that is the case, we assume the generator works correctly
                # and *all* of ``conn.response`` have been yielded already.
                if last_entry is None or last_entry not in conn.response:
                    for entry in conn.response:
                        yield entry
                break
            else:
                raise


def cache(func):
    """
    cache the user with his loginname, resolver and UID in a local
    dictionary cache.
    This is a per process cache.
    """
    @functools.wraps(func)
    def cache_wrapper(self, *args, **kwds):
        # Only run the code, in case we have a configured cache!
        if self.cache_timeout > 0:
            # If it does not exist, create the node for this instance
            resolver_id = self.getResolverId()
            now = datetime.datetime.now(tz=datetime.timezone.utc)
            tdelta = datetime.timedelta(seconds=self.cache_timeout)
            if resolver_id not in CACHE:
                CACHE[resolver_id] = {"getUserId": {},
                                      "getUserInfo": {},
                                      "_getDN": {}}
            else:
                # Clean up the cache in the current resolver and the current function
                _to_be_deleted = []
                try:
                    for user, cached_result in CACHE[resolver_id].get(func.__name__).items():
                        if now > cached_result.get("timestamp") + tdelta:
                            _to_be_deleted.append(user)
                except RuntimeError:
                    # This might happen if thread A evicts an expired
                    # cache entry while thread B looks for expired cache entries
                    pass
                for user in _to_be_deleted:
                    try:
                        del CACHE[resolver_id][func.__name__][user]
                    except KeyError:
                        pass
                del _to_be_deleted

            # get the portion of the cache for this very LDAP resolver
            r_cache = CACHE.get(resolver_id).get(func.__name__)
            entry = r_cache.get(args[0])
            if entry and now < entry.get("timestamp") + tdelta:
                log.debug("Reading {0!r} from cache for {1!r}".format(args[0], func.__name__))
                return entry.get("value")

        f_result = func(self, *args, **kwds)

        if self.cache_timeout > 0:
            # now we cache the result
            CACHE[resolver_id][func.__name__][args[0]] = {
                "value": f_result,
                "timestamp": now}

        return f_result

    return cache_wrapper


class AUTHTYPE(object):
    SIMPLE = "Simple"
    SASL_DIGEST_MD5 = "SASL Digest-MD5"
    NTLM = "NTLM"
    SASL_KERBEROS = "SASL Kerberos"


class IdResolver(UserIdResolver):
    # If the resolver could be configured editable
    updateable = True

    def __init__(self):
        self.i_am_bound = False
        self.uri = ""
        self.basedn = ""
        self.binddn = ""
        self.bindpw = ""
        self.object_classes = []
        self.dn_template = ""
        self.timeout = 5  # seconds!
        self.sizelimit = 500
        self.loginname_attribute = [""]
        self.searchfilter = ""
        self.userinfo = {}
        self.multivalueattributes = []
        self.uidtype = ""
        self.noreferrals = False
        self._editable = False
        self.resolverId = self.uri
        self.scope = ldap3.SUBTREE
        self.cache_timeout = 120
        self.tls_context = None
        self.start_tls = False
        self.serverpool_persistent = False
        self.serverpool_rounds = SERVERPOOL_ROUNDS
        self.serverpool_skip = SERVERPOOL_SKIP
        self.serverpool_strategy = SERVERPOOL_STRATEGY
        self.serverpool = None
        self.keytabfile = None
        # The number of seconds that ldap3 waits if no server is left in the pool, before
        # starting the next round
        pooling_loop_timeout = get_app_config_value("PI_LDAP_POOLING_LOOP_TIMEOUT", 10)
        log.debug("Setting system wide POOLING_LOOP_TIMEOUT to {0!s}.".format(pooling_loop_timeout))
        ldap3.set_config_parameter("POOLING_LOOP_TIMEOUT", pooling_loop_timeout)

    @log_with(log, hide_args=[2])
    def checkPass(self, uid, password):
        """
        This function checks the password for a given uid.
        - returns true in case of success
        -         false if password does not match

        """
        if self.authtype == AUTHTYPE.SASL_KERBEROS:
            if not have_gssapi:
                log.warning('gssapi module not available. Kerberos authentication not possible')
                return False
            # We need to check credentials with kerberos differently since we
            # can not use bind for every user
            upn = self.getUserInfo(uid).get('upn')
            if upn is not None and upn != "None" and upn != "":
                name = gssapi.Name(upn.upper())
            else:
                name = gssapi.Name(self.getUserInfo(uid).get('username'))
            try:
                gssapi.raw.ext_password.acquire_cred_with_password(name, to_bytes(password))
            except gssapi.exceptions.GSSError as e:
                log.info('Failed to authenticate user {0!s} with GSSAPI: {1!r}'.format(name, e))
                log.debug(traceback.format_exc())
                return False
            return True
        elif self.authtype == AUTHTYPE.NTLM:  # pragma: no cover
            # fetch the PreWindows 2000 Domain from the self.binddn
            # which would be of the format DOMAIN\username and compose the
            # bind_user to DOMAIN\sAMAccountName
            domain_name = self.binddn.split('\\')[0]
            uinfo = self.getUserInfo(uid)
            # In fact, we need the sAMAccountName. If the username mapping is
            # another attribute than the sAMAccountName the authentication
            # will fail!
            bind_user = "{0!s}\\{1!s}".format(domain_name, uinfo.get("username"))
        else:
            bind_user = self._getDN(uid)

        if not self.serverpool:
            self.serverpool = self.get_serverpool_instance(get_info=ldap3.NONE)

        try:
            log.debug("Authtype: {0!r}".format(self.authtype))
            log.debug("user    : {0!r}".format(bind_user))
            # Whatever happens. If we have an empty bind_user, we must break
            # since we must avoid anonymous binds!
            if not bind_user or len(bind_user) < 1:
                raise ResolverError("No valid user. Empty bind_user.")
            connection = self.create_connection(authtype=self.authtype,
                                                server=self.serverpool,
                                                user=bind_user,
                                                password=password,
                                                receive_timeout=self.timeout,
                                                auto_referrals=not self.noreferrals,
                                                start_tls=self.start_tls)
            if not connection.bind():
                raise ResolverError(f"Bind failed with: {connection.result.get('description')} "
                                    f"({connection.result.get('result')})")
            log.debug(f"LDAP bind operation took {connection.usage.elapsed_time}")
            connection.unbind()
            log.debug("unbind successful.")
        except Exception as e:
            log.info(f"Failed to check password for {uid!r}/{bind_user!r}: {e}")
            log.debug(traceback.format_exc())
            return False

        return True

    def _trim_result(self, result_list):
        """
        The resultlist can contain entries of type:searchResEntry and of
        type:searchResRef. If self.noreferrals is true, all type:searchResRef
        will be removed.

        :param result_list: The result list of a LDAP search
        :type result_list: resultlist (list of dicts)
        :return: new resultlist
        """
        if self.noreferrals:
            new_list = []
            for result in result_list:
                if result.get("type") == "searchResEntry":
                    new_list.append(result)
                elif result.get("type") == "searchResRef":
                    # This is a Referral
                    pass
        else:
            new_list = result_list

        return new_list

    @staticmethod
    def _escape_loginname(loginname):
        """
        This function escapes the loginname according to
        https://msdn.microsoft.com/en-us/library/aa746475(v=vs.85).aspx
        This is to avoid username guessing by trying to login as user
           a*
           ac*
           ach*
           achm*
           achme*
           achemd*

        :param loginname: The loginname
        :return: The escaped loginname
        """
        return loginname.replace("\\", "\\5c").replace("*", "\\2a").replace(
            "(", "\\28").replace(")", "\\29").replace("/", "\\2f")

    @staticmethod
    def _get_uid(entry, uidtype):
        if uidtype.lower() == "dn":
            uid = entry.get("dn")
        else:
            attributes = entry.get("attributes")
            if type(attributes.get(uidtype)) == list:
                uid = attributes.get(uidtype)[0]
            else:
                uid = attributes.get(uidtype)

        try:
            uid = convert_column_to_unicode(uid)
        except UnicodeDecodeError as e:
            # in some cases ldap3 fails to decode the uid and return it as a byte-array
            # if the utf-8 decoding fails, we try the UUID conversion
            if uidtype.lower() == "objectguid":
                # Active Directory uses little endian byte order
                log.debug(f"Found a byte-array as uid ({binascii.hexlify(uid)}), trying to convert it to a UUID "
                          f"assuming little endian byte order. ({e})")
                log.debug(traceback.format_exc())
                uid = str(uuid.UUID(bytes_le=uid))
            else:
                # ldap3 defines a standard formatter using big endian byte order for GUID (eDirectory), entryUUID
                # (openLDAP), and UUID. Hence, we assume it as the default byte order here.
                log.debug(f"Found a byte-array as uid ({binascii.hexlify(uid)}), trying to convert it to a UUID "
                          f"assuming big endian byte order. ({e})")
                log.debug(traceback.format_exc())
                uid = str(uuid.UUID(bytes=uid))

        if uidtype.lower() == "objectguid":
            # For ldap3 versions <= 2.4.1, objectGUID attribute values are returned as UUID strings.
            # For versions greater than 2.4.1, they are returned in the curly-braced string
            # representation, i.e. objectGUID := "{" UUID "}"
            # In order to ensure backwards compatibility for user mappings,
            # we strip the curly braces from objectGUID values.
            # If we are using ldap3 <= 2.4.1, there are no curly braces and we leave the value unchanged.
            uid = uid.strip("{").strip("}")

        return uid

    def _trim_user_id(self, user_id):
        """
        If we search for the objectGUID we can not search for the normal
        string representation, but we need to search for the bytestring in AD.

        :param user_id: The userId
        :return: the trimmed userId
        """
        if self.uidtype == "objectGUID":
            user_id = trim_objectGUID(user_id)
        return user_id

    @cache
    def _getDN(self, user_id):
        """
        This function returns the DN of a userId.
        Therefor it evaluates the self.uidtype.

        :param user_id: The userid of a user
        :type user_id: string
        :return: The DN of the object.
        """
        dn = ""
        if self.uidtype.lower() == "dn":
            dn = user_id
        else:
            # get the DN for the Object
            search_userid = self._trim_user_id(user_id)
            search_filter = f"(&{self.searchfilter}({self.uidtype}={search_userid}))"
            result = self._search(search_base=self.basedn, search_filter=search_filter,
                                  attributes=list(self.userinfo.values()))
            if len(result) > 1:  # pragma: no cover
                raise ResolverError(f"Found more than one object for uid {user_id!r}")
            elif len(result) == 1:
                dn = result[0].get("dn")
            else:
                log.info(f"The filter '{search_filter}' returned no DN.")
        return dn

    def _bind(self):
        """
        Perform LDAP bind operation on a connection.
        Create the connection if it doesn't exist yet
        """
        if not self.i_am_bound:
            if not self.serverpool:
                self.serverpool = self.get_serverpool_instance(self.get_info)
            try:
                self.connection = self.create_connection(authtype=self.authtype,
                                                         server=self.serverpool,
                                                         user=self.binddn,
                                                         password=self.bindpw,
                                                         receive_timeout=self.timeout,
                                                         auto_referrals=not
                                                         self.noreferrals,
                                                         start_tls=self.start_tls,
                                                         keytabfile=self.keytabfile)
                bound = self.connection.bind()
            except Exception as ex:
                log.error(f"Error performing bind operation: {ex}!")
                raise ResolverError(f"Error performing bind operation: {ex}")
            if not bound:
                result = self.connection.result
                log.error(f"LDAP Bind unsuccessful: "
                          f"{result.get('description')} ({result.get('result')})!")
                raise ResolverError(f"Unable to perform bind operation: "
                                    f"{result.get('description')} ({result.get('result')})")
            self.i_am_bound = True

    def _search(self, search_base, search_filter, attributes):
        self._bind()
        self.connection.search(search_base=search_base,
                               search_scope=self.scope,
                               search_filter=search_filter,
                               attributes=attributes)
        result = self.connection.response
        result = self._trim_result(result)
        log.debug(f"LDAP search operation took {self.connection.usage.elapsed_time}")
        return result

    @staticmethod
    def _get_tls_context(ldap_uri=None, start_tls=False, tls_version=None, tls_verify=None,
                         tls_ca_file=None, tls_options=None):
        """
        This method creates the Tls object to be used with ldap3.
        """
        if ldap_uri.lower().startswith("ldaps") or is_true(start_tls):
            if not tls_version:
                tls_version = int(DEFAULT_TLS_PROTOCOL)
            # If TLS_VERSION is 2, set tls_options to use TLS v1.3
            if not tls_options:
                tls_options = TLS_OPTIONS_1_3 if int(tls_version) == int(TLS_NEGOTIATE_PROTOCOL) else None
            if tls_verify:
                tls_ca_file = tls_ca_file or DEFAULT_CA_FILE
            else:
                tls_verify = ssl.CERT_NONE
                tls_ca_file = None
            tls_context = Tls(validate=tls_verify,
                              version=int(tls_version),
                              ssl_options=tls_options,
                              ca_certs_file=tls_ca_file)
        else:
            tls_context = None

        return tls_context

    @cache
    def getUserInfo(self, user_id):
        """
        This function returns all user info for a given userid/object.

        :param user_id: The userid of the object
        :type user_id: string
        :return: A dictionary with the keys defined in self.userinfo
        :rtype: dict
        """
        user_info = {}

        if self.uidtype.lower() == "dn":
            # encode utf8, so that also german umlauts work in the DN
            search_filter = f"(&{self.searchfilter})"
            search_base = user_id
        else:
            search_uid = to_unicode(self._trim_user_id(user_id))
            search_filter = f"(&{self.searchfilter}({self.uidtype}={search_uid}))"
            search_base = self.basedn

        result = self._search(search_base=search_base, search_filter=search_filter,
                              attributes=list(self.userinfo.values()))

        if len(result) > 1:  # pragma: no cover
            raise ResolverError(f"Found more than one object for uid {user_id!r}")

        for entry in result:
            attributes = entry.get("attributes")
            user_info = self._ldap_attributes_to_user_object(attributes)

        return user_info

    def _ldap_attributes_to_user_object(self, attributes):
        """
        This helper function converts the LDAP attributes to a dictionary for
        the privacyIDEA user. The LDAP Userinfo mapping is used to do so.

        :param attributes:
        :return: dict with privacyIDEA user info.
        :rtype: dict
        """
        user_info = {}
        for ldap_k, ldap_v in attributes.items():
            for map_k, map_v in self.userinfo.items():
                if ldap_k == map_v:
                    if ldap_k == "objectGUID":
                        # An objectGUID should be no list, since it is unique
                        if isinstance(ldap_v, str):
                            user_info[map_k] = ldap_v.strip("{").strip("}")
                        else:
                            raise Exception("The LDAP returns an objectGUID, "
                                            "that is no string: {0!s}".format(type(ldap_v)))
                    elif isinstance(ldap_v, list) and map_k not in self.multivalueattributes:
                        # lists that are not in self.multivalueattributes return first value
                        # as a string. Multi-value-attributes are returned as a list
                        if ldap_v:
                            user_info[map_k] = ldap_v[0]
                        else:
                            user_info[map_k] = ""
                    else:
                        user_info[map_k] = ldap_v
        return user_info

    def getUsername(self, user_id):
        """
        Returns the username/loginname for a given user_id

        :param user_id: The user_id in this resolver
        :type user_id: string
        :return: username
        :rtype: string
        """
        info = self.getUserInfo(user_id)
        return info.get('username', "")

    @cache
    def getUserId(self, login_name):
        """
        resolve the loginname to the userid.

        :param login_name: The login name from the credentials
        :type login_name: str
        :return: UserId as found for the LoginName
        :rtype: str
        """
        userid = ""
        login_name = to_unicode(login_name)
        login_name = self._escape_loginname(login_name)

        if len(self.loginname_attribute) > 1:
            loginname_filter = ""
            for l_attribute in self.loginname_attribute:
                # Special case if we have a guid
                try:
                    if l_attribute.lower() == "objectguid":
                        search_login_name = trim_objectGUID(login_name)
                    else:
                        search_login_name = login_name
                    loginname_filter += "({!s}={!s})".format(l_attribute.strip(),
                                                             search_login_name)
                except ValueError:
                    # This happens if we have a self.loginname_attribute like ["sAMAccountName","objectGUID"],
                    # the user logs in with his sAMAccountName, which can
                    # not be transformed to a UUID
                    log.debug("Can not transform {0!s} to a objectGUID.".format(login_name))

            loginname_filter = "|" + loginname_filter
        else:
            if self.loginname_attribute[0].lower() == "objectguid":
                search_login_name = trim_objectGUID(login_name)
            else:
                search_login_name = login_name
            loginname_filter = "{!s}={!s}".format(self.loginname_attribute[0],
                                                  search_login_name)

        log.debug("login name filter: {!r}".format(loginname_filter))
        search_filter = f"(&{self.searchfilter}({loginname_filter}))"

        # create search attributes
        attributes = list(self.userinfo.values())
        if self.uidtype.lower() != "dn":
            attributes.append(str(self.uidtype))

        log.debug("Searching user {0!r} in LDAP.".format(login_name))
        result = self._search(search_base=self.basedn, search_filter=search_filter,
                              attributes=attributes)
        if len(result) > 1:  # pragma: no cover
            raise ResolverError(f"Found more than one object for Login {login_name!r}")

        for entry in result:
            userid = self._get_uid(entry, self.uidtype)

        return userid

    def getUserList(self, search_dict=None):
        """
        :param search_dict: A dictionary with search parameters
        :type search_dict: dict
        :return: list of users, where each user is a dictionary
        """
        user_list = []
        attributes = list(self.userinfo.values())
        if self.uidtype.lower() != "dn":
            attributes.append(str(self.uidtype))

        # do the filter depending on the searchDict
        search_filter = "(&" + self.searchfilter
        for search_key in search_dict.keys():
            # convert to unicode
            search_dict[search_key] = to_unicode(search_dict[search_key])
            if search_key == "accountExpires":
                comperator = ">="
                if search_dict[search_key] in ["1", 1]:
                    comperator = "<="
                search_filter += "(&({0!s}{1!s}{2!s})(!({3!s}=0)))".format(
                    self.userinfo[search_key], comperator,
                    get_ad_timestamp_now(), self.userinfo[search_key])
            else:
                search_filter += "({0!s}={1!s})".format(self.userinfo[search_key],
                                                        search_dict[search_key])
        search_filter += ")"

        self._bind()
        try:
            search_generator = self.connection.extend.standard.paged_search(search_base=self.basedn,
                                                                            search_filter=search_filter,
                                                                            search_scope=self.scope,
                                                                            attributes=attributes,
                                                                            paged_size=100,
                                                                            size_limit=self.sizelimit,
                                                                            generator=True)
            log.debug(f"LDAP paged search operation took {self.connection.usage.elapsed_time}")
        except Exception as e:
            log.error(f"Error performing paged search: {e}")
            raise ResolverError(f"Error performing paged search: {e}")
        # returns a generator of dictionaries
        for entry in ignore_sizelimit_exception(self.connection, search_generator):
            # Simple fix for ignored sizelimit with Active Directory
            if len(user_list) >= self.sizelimit:
                break
            # Fix for searchResRef entries which have no attributes
            if entry.get('type') == 'searchResRef':
                continue
            try:
                attributes = entry.get("attributes")
                user = self._ldap_attributes_to_user_object(attributes)
                user['userid'] = self._get_uid(entry, self.uidtype)
                user_list.append(user)
            except Exception as ex:  # pragma: no cover
                log.error(f"Error during fetching LDAP objects: {ex}")
                log.debug("{0!s}".format(traceback.format_exc()))

        return user_list

    def getResolverId(self):
        """
        Returns the resolver Id
        This should be an Identifier of the resolver, preferable the type
        and the name of the resolver.

        :return: the id of the resolver
        :rtype: str
        """
        id_string = (f"{self.uri}{self.basedn}{self.searchfilter}"
                     f"{sorted(self.userinfo.items(), key=itemgetter(0))}")
        result = binascii.hexlify(hashlib.sha1(id_string.encode("utf-8")).digest())  # nosec B324 # hash used as unique identifier
        return result.decode('utf8')

    @staticmethod
    def getResolverClassType():
        return 'ldapresolver'

    @staticmethod
    def getResolverDescriptor():
        return IdResolver.getResolverClassDescriptor()

    @staticmethod
    def getResolverType():
        return IdResolver.getResolverClassType()

    def loadConfig(self, config):
        """
        Load the config from conf.

        :param config: The configuration from the Config Table
        :type config: dict


        '#ldap_uri': 'LDAPURI',
        '#ldap_basedn': 'LDAPBASE',
        '#ldap_binddn': 'BINDDN',
        '#ldap_password': 'BINDPW',
        '#ldap_timeout': 'TIMEOUT',
        '#ldap_sizelimit': 'SIZELIMIT',
        '#ldap_loginattr': 'LOGINNAMEATTRIBUTE',
        '#ldap_searchfilter': 'LDAPSEARCHFILTER',
        '#ldap_mapping': 'USERINFO',
        '#ldap_uidtype': 'UIDTYPE',
        '#ldap_noreferrals' : 'NOREFERRALS',
        '#ldap_editable' : 'EDITABLE',
        '#ldap_certificate': 'CACERTIFICATE',
        '#ldap_keytabfile': 'KEYTABFILE',

        """
        self.uri = config.get("LDAPURI")
        self.basedn = config.get("LDAPBASE")
        self.binddn = config.get("BINDDN")
        # object_classes is a comma separated list like
        # ["top", "person", "organizationalPerson", "user", "inetOrgPerson"]
        self.object_classes = [cl.strip() for cl in config.get("OBJECT_CLASSES", "").split(",")]
        self.dn_template = config.get("DN_TEMPLATE", "")
        self.bindpw = config.get("BINDPW")
        self.timeout = int(config.get("TIMEOUT", 5))
        self.cache_timeout = int(config.get("CACHE_TIMEOUT", 120))
        self.sizelimit = int(config.get("SIZELIMIT", 500))
        self.loginname_attribute = [la.strip() for la in config.get("LOGINNAMEATTRIBUTE", "").split(",")]
        self.searchfilter = config.get("LDAPSEARCHFILTER")
        userinfo = config.get("USERINFO", "{}")
        self.userinfo = yaml.safe_load(userinfo)
        self.userinfo["username"] = self.loginname_attribute[0]
        multivalueattributes = config.get("MULTIVALUEATTRIBUTES") or '["mobile"]'
        self.multivalueattributes = yaml.safe_load(multivalueattributes)
        self.map = yaml.safe_load(userinfo)
        self.uidtype = config.get("UIDTYPE", "DN")
        self.noreferrals = is_true(config.get("NOREFERRALS", False))
        self.start_tls = is_true(config.get("START_TLS", False)) and not self.uri.lower().startswith("ldaps")
        self.get_info = get_info_configuration(is_true(config.get("NOSCHEMAS", False)))
        self._editable = config.get("EDITABLE", False)
        self.scope = config.get("SCOPE") or ldap3.SUBTREE
        self.resolverId = self.uri
        self.authtype = config.get("AUTHTYPE", AUTHTYPE.SIMPLE)
        self.keytabfile = config.get('KEYTABFILE', None)
        self.tls_verify = is_true(config.get("TLS_VERIFY", False))
        # Fallback to DEFAULT_TLS_PROTOCOL (TLSv1: 3, TLSv1.1: 4, v1.2: 5, TLS negotiation: 2)
        self.tls_version = int(config.get("TLS_VERSION") or DEFAULT_TLS_PROTOCOL)
        self.tls_ca_file = config.get("TLS_CA_FILE")
        self.tls_context = self._get_tls_context(ldap_uri=self.uri, start_tls=self.start_tls,
                                                 tls_version=self.tls_version,
                                                 tls_verify=self.tls_verify,
                                                 tls_ca_file=self.tls_ca_file)
        self.serverpool_persistent = is_true(config.get("SERVERPOOL_PERSISTENT", False))
        self.serverpool_rounds = int(config.get("SERVERPOOL_ROUNDS") or SERVERPOOL_ROUNDS)
        self.serverpool_skip = int(config.get("SERVERPOOL_SKIP") or SERVERPOOL_SKIP)
        self.serverpool_strategy = config.get("SERVERPOOL_STRATEGY") or SERVERPOOL_STRATEGY
        # The configuration might have changed. We reset the serverpool
        self.serverpool = None
        self.i_am_bound = False

        return self

    @property
    def has_multiple_loginnames(self):
        """
        Return if this resolver has multiple loginname attributes
        :return: bool
        """
        return len(self.loginname_attribute) > 1

    @staticmethod
    def split_uri(uri):
        """
        Splits LDAP URIs like:
        * ldap://server
        * ldaps://server
        * ldap[s]://server:1234
        * server
        :param uri: The LDAP URI
        :return: Returns a tuple of Servername, Port and SSL(bool)
        """
        port = None
        use_ssl = False
        ldap_elems = uri.split(":")
        if len(ldap_elems) == 3:
            server = ldap_elems[1].strip("/")
            port = int(ldap_elems[2])
            if ldap_elems[0].lower() == "ldaps":
                use_ssl = True
            else:
                use_ssl = False
        elif len(ldap_elems) == 2:
            server = ldap_elems[1].strip("/")
            port = None
            if ldap_elems[0].lower() == "ldaps":
                use_ssl = True
            else:
                use_ssl = False
        else:
            server = uri

        return server, port, use_ssl

    @classmethod
    def create_serverpool(cls, urilist, timeout, get_info=None, tls_context=None, rounds=SERVERPOOL_ROUNDS,
                          exhaust=SERVERPOOL_SKIP, pool_cls=ldap3.ServerPool, strategy=SERVERPOOL_STRATEGY):
        """
        This creates the serverpool for the ldap3 connection.
        The URI from the LDAP resolver can contain a comma separated list of
        LDAP servers. These are split and then added to the pool.

        See
        https://github.com/cannatag/ldap3/blob/master/docs/manual/source/servers.rst#server-pool

        :param urilist: The list of LDAP URIs, comma separated
        :type urilist: basestring
        :param timeout: The connection timeout
        :type timeout: float
        :param get_info: The get_info type passed to the ldap3.Sever
            constructor. default: ldap3.SCHEMA, should be ldap3.NONE in case
            of a bind.
        :param tls_context: A ldap3.tls object, which defines if certificate
            verification should be performed
        :param rounds: The number of rounds we should cycle through the server pool
            before giving up
        :param exhaust: The seconds, for how long a non-reachable server should be
            removed from the serverpool
        :param pool_cls: ``ldap3.ServerPool`` subclass that should be instantiated
        :param strategy: The pooling strategy of the server-pool
        :type strategy: str
        :return: Server Pool
        :rtype: ldap3.ServerPool
        """
        get_info = get_info or ldap3.SCHEMA
        strategy = LDAP_STRATEGY.get(strategy, SERVERPOOL_STRATEGY)
        server_pool = pool_cls(None, strategy,
                               active=rounds,
                               exhaust=exhaust)
        for uri in urilist.split(","):
            uri = uri.strip()
            host, port, ssl = cls.split_uri(uri)
            server = ldap3.Server(host, port=port,
                                  use_ssl=ssl,
                                  connect_timeout=float(timeout),
                                  get_info=get_info,
                                  tls=tls_context)
            server_pool.add(server)
            log.debug("Added {0!s}, {1!s}, {2!s} to server pool.".format(host, port, ssl))
        return server_pool

    def get_serverpool_instance(self, get_info=None):
        """
        Return a ``ServerPool`` instance that should be used. If ``SERVERPOOL_PERSISTENT``
        is enabled, invoke ``get_persistent_serverpool`` to retrieve a per-process
        server pool instance. If it is not enabled, invoke ``create_serverpool``
        to retrieve a per-request server pool instance.

        :param get_info: one of ldap3.SCHEMA, ldap3.NONE, ldap3.ALL
        :return: a ``ServerPool``/``LockingServerPool`` instance
        """
        if self.serverpool_persistent:
            return self.get_persistent_serverpool(get_info)
        else:
            return self.create_serverpool(self.uri, self.timeout, get_info, self.tls_context,
                                          self.serverpool_rounds, self.serverpool_skip,
                                          strategy=self.serverpool_strategy)

    def get_persistent_serverpool(self, get_info=None):
        """
        Return a process-level instance of ``LockingServerPool`` for the current LDAP resolver
        configuration. Retrieve it from the app-local store. If such an instance does not exist
        yet, create one.

        :param get_info: one of ldap3.SCHEMA, ldap3.NONE, ldap3.ALL
        :return: a ``LockingServerPool`` instance
        """
        if not get_info:
            get_info = ldap3.SCHEMA
        pools = get_app_local_store().setdefault('ldap_server_pools', {})
        # Create a hashable tuple that describes the current server pool configuration
        pool_description = (self.uri,
                            self.timeout,
                            get_info,
                            repr(self.tls_context),  # this is the string representation of the TLS context
                            self.serverpool_rounds,
                            self.serverpool_skip)
        if pool_description not in pools:
            log.debug("Creating a persistent server pool instance for {!r} ...".format(pool_description))
            # Create a suitable instance of ``LockingServerPool``
            server_pool = self.create_serverpool(self.uri, self.timeout, get_info,
                                                 self.tls_context, self.serverpool_rounds, self.serverpool_skip,
                                                 pool_cls=LockingServerPool, strategy=self.serverpool_strategy)
            # It may happen that another thread tries to add an instance to the dictionary concurrently.
            # However, only one of them will win, and the other ``LockingServerPool`` instance will be
            # garbage-collected eventually.
            return pools.setdefault(pool_description, server_pool)
        else:
            # If there is already a ``LockingServerPool`` instance, return it.
            # We never remove instances from the dictionary, so a ``KeyError`` cannot occur.
            # As a side effect, when we change the LDAP Id resolver configuration,
            # outdated ``LockingServerPool`` instances will survive until the next server restart.
            return pools[pool_description]

    @classmethod
    def getResolverClassDescriptor(cls):
        """
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        """
        descriptor = {}
        typ = cls.getResolverType()
        descriptor['clazz'] = "useridresolver.LDAPIdResolver.IdResolver"
        descriptor['config'] = {'LDAPURI': 'string',
                                'LDAPBASE': 'string',
                                'BINDDN': 'string',
                                'BINDPW': 'password',
                                'KEYTABFILE': 'string',
                                'TIMEOUT': 'int',
                                'SIZELIMIT': 'int',
                                'LOGINNAMEATTRIBUTE': 'string',
                                'LDAPSEARCHFILTER': 'string',
                                'USERINFO': 'string',
                                'UIDTYPE': 'string',
                                'NOREFERRALS': 'bool',
                                'NOSCHEMAS': 'bool',
                                'CACERTIFICATE': 'string',
                                'EDITABLE': 'bool',
                                'SCOPE': 'string',
                                'AUTHTYPE': 'string',
                                'TLS_VERIFY': 'bool',
                                'TLS_VERSION': 'int',
                                'TLS_CA_FILE': 'string',
                                'START_TLS': 'bool',
                                'CACHE_TIMEOUT': 'int',
                                'SERVERPOOL_STRATEGY': 'string',
                                'SERVERPOOL_ROUNDS': 'int',
                                'SERVERPOOL_SKIP': 'int',
                                'SERVERPOOL_PERSISTENT': 'bool',
                                'OBJECT_CLASSES': 'string',
                                'DN_TEMPLATE': 'string',
                                'MULTIVALUEATTRIBUTES': 'string'}
        return {typ: descriptor}

    @classmethod
    def testconnection(cls, param):
        """
        This function lets you test the to be saved LDAP connection.

        Parameters are:
            BINDDN, BINDPW, LDAPURI, TIMEOUT, LDAPBASE, LOGINNAMEATTRIBUTE,
            LDAPSEARCHFILTER, USERINFO, SIZELIMIT, NOREFERRALS, CACERTIFICATE,
            AUTHTYPE, TLS_VERIFY, TLS_VERSION, TLS_CA_FILE, SERVERPOOL_ROUNDS,
            SERVERPOOL_SKIP, SERVERPOOL_STRATEGY

        :param param: A dictionary with all necessary parameter to test
                        the connection.
        :type param: dict
        :return: Tuple of success and a description
        :rtype: (bool, string)
        """
        success = False
        uidtype = param.get("UIDTYPE")
        timeout = int(param.get("TIMEOUT", 5))
        ldap_uri = param.get("LDAPURI")
        size_limit = int(param.get("SIZELIMIT", 500))
        serverpool_rounds = int(param.get("SERVERPOOL_ROUNDS") or SERVERPOOL_ROUNDS)
        serverpool_skip = int(param.get("SERVERPOOL_SKIP") or SERVERPOOL_SKIP)
        pool_strat = param.get("SERVERPOOL_STRATEGY") or SERVERPOOL_STRATEGY
        serverpool_strategy = LDAP_STRATEGY.get(pool_strat, SERVERPOOL_STRATEGY)
        start_tls = is_true(param.get("START_TLS", False)) and not ldap_uri.lower().startswith("ldaps")
        tls_verify = is_true(param.get("TLS_VERIFY"))
        tls_context = cls._get_tls_context(ldap_uri=ldap_uri,
                                           start_tls=start_tls,
                                           tls_version=param.get("TLS_VERSION"),
                                           tls_verify=tls_verify,
                                           tls_ca_file=param.get("TLS_CA_FILE"),
                                           tls_options=None)
        get_info = get_info_configuration(is_true(param.get("NOSCHEMAS")))
        try:
            server_pool = cls.create_serverpool(ldap_uri, timeout,
                                                tls_context=tls_context,
                                                get_info=get_info,
                                                rounds=serverpool_rounds,
                                                exhaust=serverpool_skip,
                                                strategy=serverpool_strategy)
            conn = cls.create_connection(authtype=param.get("AUTHTYPE", AUTHTYPE.SIMPLE),
                                         server=server_pool,
                                         user=param.get("BINDDN"),
                                         password=param.get("BINDPW"),
                                         receive_timeout=timeout,
                                         auto_referrals=not param.get("NOREFERRALS"),
                                         start_tls=start_tls,
                                         keytabfile=param.get('KEYTABFILE', None))

            if not conn.bind():
                log.error(f"LDAP Bind unsuccessful: "
                          f"{conn.result.get('description')} ({conn.result.get('result')})!")
                raise ResolverError(f"Unable to perform bind operation: "
                                    f"{conn.result.get('description')} ({conn.result.get('result')})!")
            # create searchattributes
            attributes = list(yaml.safe_load(param["USERINFO"]).values())
            if uidtype.lower() != "dn":
                attributes.append(str(uidtype))
            # search for users...
            search_generator = conn.extend.standard.paged_search(
                search_base=param["LDAPBASE"],
                search_filter="(&" + param["LDAPSEARCHFILTER"] + ")",
                search_scope=param.get("SCOPE") or ldap3.SUBTREE,
                attributes=attributes,
                paged_size=100,
                size_limit=size_limit,
                generator=True)
            # returns a generator of dictionaries
            elapsed_time = conn.usage.elapsed_time.total_seconds()
            count = 0
            uidtype_count = 0
            for entry in ignore_sizelimit_exception(conn, search_generator):
                # Fix for searchResRef entries which have no attributes
                if entry.get('type') == 'searchResRef':
                    continue
                try:
                    userid = cls._get_uid(entry, uidtype)
                    count += 1
                    if userid:
                        uidtype_count += 1
                except Exception as exx:  # pragma: no cover
                    log.warning("Error during fetching LDAP objects:"
                                " {0!r}".format(exx))
                    log.debug("{0!s}".format(traceback.format_exc()))

            if uidtype_count < count:  # pragma: no cover
                message = _("Your LDAP config found {0!s} user objects in {2:.4f}s, "
                            "but only {1!s} with the specified "
                            "uidtype.").format(count, uidtype_count, elapsed_time)
            else:
                message = _("Your LDAP config seems to be OK, {0!s} user objects "
                            "found in {1:.4f}s.").format(count, elapsed_time)

            conn.unbind()
            success = True

        except Exception as e:
            message = f"{e}"
            log.debug("{0!s}".format(traceback.format_exc()))

        return success, message

    def add_user(self, attributes=None):
        """
        Add a new user to the LDAP directory.
        The user can only be created in the LDAP using a DN.
        So we have to construct the DN out of the given attributes.

        attributes are these
        "username", "surname", "givenname", "email",
        "mobile", "phone", "password"

        :param attributes: Attributes according to the attribute mapping
        :type attributes: dict
        :return: The new UID of the user. The UserIdResolver needs to
        determine the way how to create the UID.
        """
        # TODO: We still have some utf8 issues creating users with special characters.
        attributes = attributes or {}

        dn = self.dn_template
        dn = dn.replace("<basedn>", self.basedn)
        dn = dn.replace("<username>", attributes.get("username", ""))
        dn = dn.replace("<givenname>", attributes.get("givenname", ""))
        dn = dn.replace("<surname>", attributes.get("surname", ""))

        try:
            self._bind()
            params = self._attributes_to_ldap_attributes(attributes)
            self.connection.add(dn, self.object_classes, params)
            log.debug(f"LDAP add operation took {self.connection.usage.elapsed_time}")

        except Exception as e:
            log.warning("Error accessing LDAP server: {0!r}".format(e))
            log.debug("{0}".format(traceback.format_exc()))
            raise privacyIDEAError(e)

        if self.connection.result.get("result") != 0:
            log.warning(f"Error during adding of user {dn}: "
                        f"{self.connection.result.get('description')} ({self.connection.result.get('result')})")
            raise privacyIDEAError(self.connection.result.get('message'))

        return self.getUserId(attributes.get("username"))

    def delete_user(self, uid):
        """
        Delete a user from the LDAP Directory.
        The user is referenced by the user id.

        :param uid: The uid of the user object, that should be deleted.
        :type uid: str
        :return: Returns True in case of success
        :rtype: bool
        """
        res = True
        try:
            self._bind()

            self.connection.delete(self._getDN(uid))
            log.debug(f"LDAP delete operation took {self.connection.usage.elapsed_time}")
        except Exception as ex:
            log.warning(f"Failed to delete user with uid {uid}: {ex}")
            res = False
        return res

    def _attributes_to_ldap_attributes(self, attributes):
        """
        takes the attributes and maps them to the LDAP attributes
        :param attributes: Attributes to be updated
        :type attributes: dict
        :return: dict with attribute name as keys and values
        """
        ldap_attributes = {}
        for fieldname, value in attributes.items():
            if self.map.get(fieldname):
                if fieldname == "password":
                    # Variable value may be either a string or a list
                    # so catch the TypeError exception if we get the wrong
                    # variable type
                    try:
                        pw_hash = ldap_salted_sha1.hash(value[1][0])
                        value[1][0] = pw_hash
                        ldap_attributes[self.map.get(fieldname)] = value
                    except TypeError as _e:
                        pw_hash = ldap_salted_sha1.hash(value)
                        ldap_attributes[self.map.get(fieldname)] = pw_hash
                else:
                    ldap_attributes[self.map.get(fieldname)] = value

        return ldap_attributes

    def _create_ldap_modify_changes(self, attributes, uid):
        """
        Identifies if an LDAP attribute already exists and if the value needs to be updated, deleted or added.

        :param attributes: Attributes to be updated
        :type attributes: dict
        :param uid: The uid of the user object in the resolver
        :type uid: basestring
        :return: dict with attribute name as keys and values
        """
        modify_changes = {}
        uinfo = self.getUserInfo(uid)

        for fieldname, value in attributes.items():
            if value:
                if fieldname in uinfo:
                    modify_changes[fieldname] = [MODIFY_REPLACE, [value]]
                else:
                    modify_changes[fieldname] = [MODIFY_ADD, [value]]
            else:
                modify_changes[fieldname] = [MODIFY_DELETE, [value]]

        return modify_changes

    def update_user(self, uid, attributes=None):
        """
        Update an existing user.
        This function is also used to update the password. Since the
        attribute mapping know, which field contains the password,
        this function can also take care for password changing.

        Attributes that are not contained in the dict attributes are not
        modified.

        :param uid: The uid of the user object in the resolver.
        :type uid: basestring
        :param attributes: Attributes to be updated.
        :type attributes: dict
        :return: True in case of success
        """
        attributes = attributes or {}
        try:
            self._bind()

            mapped = self._create_ldap_modify_changes(attributes, uid)
            params = self._attributes_to_ldap_attributes(mapped)
            self.connection.modify(self._getDN(uid), params)
            log.debug(f"LDAP modify operation took {self.connection.usage.elapsed_time}")
        except Exception as e:
            log.error("Error accessing LDAP server: {0!r}".format(e))
            log.debug("{0!s}".format(traceback.format_exc()))
            return False

        if self.connection.result.get('result') != 0:
            log.error("Error during update of user {0!r}: "
                      "{1!r}".format(uid, self.connection.result.get("message")))
            return False

        return True

    @staticmethod
    def create_connection(authtype=None, server=None, user=None,
                          password=None, auto_bind=ldap3.AUTO_BIND_NONE,
                          client_strategy=ldap3.SYNC,
                          check_names=True,
                          auto_referrals=False,
                          receive_timeout=5,
                          start_tls=False,
                          keytabfile=None):
        """
        Create a connection to the LDAP server.

        :param authtype:
        :param server:
        :param user:
        :param password:
        :param auto_bind:
        :param client_strategy:
        :param check_names:
        :param auto_referrals:
        :param receive_timeout: At the moment we do not use this,
                                since receive_timeout is not supported by ldap3 < 2
        :type receive_timeout: float
        :param start_tls: Use startTLS for connection to server
        :type start_tls: bool
        :param keytabfile: Path to keytab file for service account
        :type keytabfile: str or None
        :return:
        """
        conn_opts = {'auto_bind': auto_bind,
                     'client_strategy': client_strategy,
                     'check_names': check_names,
                     'receive_timeout': receive_timeout,
                     'auto_referrals': auto_referrals,
                     'collect_usage': True}

        if not user:
            # without a user we can only use an anonymous binds
            conn_opts.update({'authentication': ldap3.ANONYMOUS})
        elif authtype == AUTHTYPE.SIMPLE:
            # SIMPLE works with passwords as UTF8 and unicode
            password = to_unicode(password)
            conn_opts.update({'user': user,
                              'password': password,
                              'authentication': ldap3.SIMPLE})
        elif authtype == AUTHTYPE.NTLM:  # pragma: no cover
            # NTLM requires the password to be unicode
            password = to_unicode(password)
            conn_opts.update({'user': user,
                              'password': password,
                              'authentication': ldap3.NTLM})
        elif authtype == AUTHTYPE.SASL_DIGEST_MD5:  # pragma: no cover
            password = to_unicode(password)
            sasl_credentials = (str(user), str(password))
            conn_opts.update({'sasl_mechanism': ldap3.DIGEST_MD5,
                              'sasl_credentials': sasl_credentials,
                              'authentication': ldap3.SASL})
        elif authtype == AUTHTYPE.SASL_KERBEROS:
            cred_store = {'client_keytab': keytabfile} if keytabfile else None
            conn_opts.update({'sasl_mechanism': ldap3.KERBEROS,
                              'authentication': ldap3.SASL,
                              'user': user,
                              'cred_store': cred_store})
        else:
            raise ResolverError(f"Authtype {authtype} not supported")

        connection = ldap3.Connection(server, **conn_opts)
        if start_tls:
            connection.open(read_server_info=False)
            log.debug("Doing start_tls")
            try:
                if not connection.start_tls(read_server_info=False):
                    err_msg = (f"Unable to create connection with StartTLS: "
                               f"{connection.result.get('description')} "
                               f"({connection.result.get('error')})")
                    log.error(err_msg)
                    raise ResolverError(err_msg)
            except Exception as e:
                err_msg = f"Unable to create connection with StartTLS: {e}"
                log.error(err_msg)
                raise ResolverError(err_msg)

        return connection

    @property
    def editable(self):
        """
        Return true, if the instance of the resolver is configured editable
        :return:
        """
        # Depending on the database this might look different
        # Usually this is "1"
        return is_true(self._editable)



#  privacyIDEA is a fork of LinOTP
#
#  2015-11-03   Cornelius Kölbel <cornelius@privacyidea.org>
#               Add memberfunction "exist"
#  2015-06-06   Cornelius Kölbel <cornelius@privacyidea.org>
#               Add the possibility to update the user data.
#  Nov 27, 2014 Cornelius Kölbel <cornelius@privacyidea.org>
#               Migration to flask
#               Rewrite of methods
#               100% test code coverage
#  May 08, 2014 Cornelius Kölbel
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# 2014-10-03 fix getUsername function
#            Cornelius Kölbel <cornelius@privcyidea.org>
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = '''There are the library functions for user functions.
It depends on the lib.resolver and lib.realm.

There are and must be no dependencies to the token functions (lib.token)
or to webservices!

This code is tested in tests/test_lib_user.py
'''

import hashlib
import logging
import traceback

from .error import UserError
from ..api.lib.utils import (getParam,
                             optional)
from .log import log_with
from .resolver import (get_resolver_object,
                       get_resolver_type)

from .realm import (get_realms, realm_is_defined,
                    get_default_realm,
                    get_realm, get_realm_id)
from .config import get_from_config, SYSCONF
from .framework import get_app_config_value
from .usercache import (user_cache, cache_username, user_init, delete_user_cache)
from privacyidea.models import CustomUserAttribute, db

log = logging.getLogger(__name__)


class User(object):
    """
    The user has the attributes
      login, realm and resolver.
    Usually a user can be found via "login@realm".

    A user object with an empty login and realm should not exist,
    whereas a user object could have an empty resolver.
    """

    # In some test case the login attribute from a not
    # initialized user is requested. This is why we need
    # these dummy class attributes.
    login = ""
    realm = ""
    resolver = ""

    # NOTE: Directly decorating the class ``User`` breaks ``isinstance`` checks,
    # which is why we have to decorate __init__
    @log_with(log)
    def __init__(self, login="", realm="", resolver="", uid=None):
        self.login = login or ""
        self.used_login = self.login
        self.realm = (realm or "").lower()
        self.realm_id = None
        if resolver == "**":
            resolver = ""
        self.resolver = resolver or ""
        self.uid = uid
        self.rtype = None
        # Hash of already checked passwords and their result. If a user has multiple token, it is not necessary to check
        # the same password multiple times. However, we can not differentiate between PIN+OTP and just PIN, so this dict
        # will have an entry for PIN+OTP (likely to fail) and then the OTP cut off to just PIN. In case the PIN was
        # given in the request, the dict will have only one entry.
        self._checked_passwords = {}
        if not self.login and not self.resolver and uid is not None:
            raise UserError("Can not create a user object from a uid without a resolver!")
        # Enrich user object with information from the userstore or from the
        # user cache
        if login or uid is not None:
            self._get_user_from_userstore()
            # Just store the resolver type
            self.rtype = get_resolver_type(self.resolver)
            # Add realm_id to User object
            self.realm_id = get_realm_id(self.realm)

    @user_cache(user_init)
    def _get_user_from_userstore(self):
        if not self.resolver:
            # set the resolver implicitly!
            self._get_resolvers()

        # Get Identifiers
        if self.resolver:
            y = get_resolver_object(self.resolver)
            if y is None:
                raise UserError("The resolver '{0!s}' does not exist!".format(
                    self.resolver))
            if self.uid is None:
                # Determine the uid
                self.uid = y.getUserId(self.login)
            if not self.login:
                # Determine the login if it does not exist or
                self.used_login = self.login = y.getUsername(self.uid)
            if y.has_multiple_loginnames:
                # if the resolver has multiple logins the primary login might be another value!
                self.login = y.getUsername(self.uid)

    def is_empty(self):
        # ignore if only resolver is set! as it makes no sense
        if len(self.login or "") + len(self.realm or "") == 0:
            return True
        else:
            return False

    def __eq__(self, other):
        """
        Compare two User Objects.

        :param other: The other User object, to which this very object is
        compared.
        :type other: User object
        :return: True or False
        :rtype: bool
        """
        if not isinstance(other, type(self)):
            log.info("Comparing a non-user object: {0!s} != {1!s}.".format(self, type(other)))
            return False
        if (self.resolver != other.resolver) or (self.realm != other.realm):
            log.info("Users are not in the same resolver and realm: "
                     "{0!s} != {1!s}.".format(self, other))
            return False
        if self.uid and other.uid:
            log.debug("Comparing based on uid: {0!s} vs {1!s}".format(self.uid, other.uid))
            return self.uid == other.uid
        log.debug("Comparing based on login: {0!s} vs {1!s}".format(self.login, other.login))
        return self.login == other.login

    def __ne__(self, other):
        """
        Compare two user objects and return true, if they are not equal

        :param other: The other User object
        :return: True or False
        """
        return not self.__eq__(other)

    def __hash__(self):
        return hash((type(self), self.login, self.resolver, self.realm))

    def __str__(self):
        ret = "<empty user>"
        if not self.is_empty():
            # Realm and resolver should always be ASCII
            conf = ''
            if self.resolver:
                conf = '.{0!s}'.format(self.resolver)
            ret = '<{0!s}{1!s}@{2!s}>'.format(self.login, conf, self.realm)
        return ret

    def __repr__(self):
        ret = ("User(login={0!r}, realm={1!r}, resolver={2!r})".format(
            self.login, self.realm, self.resolver))
        return ret

    def __bool__(self):
        return not self.is_empty()

    __nonzero__ = __bool__

    @log_with(log)
    def get_ordered_resolvers(self):
        """
        returns a list of resolver names ordered by priority.
        The resolver with the lowest priority is the first.
        If resolvers have the same priority, they are ordered alphabetically.

        :return: list of resolver names
        :rtype: list
        """
        resolver_tuples = []
        realm_config = get_realms(self.realm)
        resolvers_in_realm = realm_config.get(self.realm, {}).get("resolver", [])
        for resolver in resolvers_in_realm:
            # append a tuple
            resolver_tuples.append((resolver.get("name"),
                                    resolver.get("priority") or 1000,
                                    resolver.get("node")))

        # sort the resolvers by the 2nd entry in the tuple, the priority
        sorted_resolvers = sorted(resolver_tuples, key=lambda res: res[1])
        # if the resolver contains a node setting, we only add it if it is on the correct node
        local_node_uuid = get_app_config_value("PI_NODE_UUID")
        resolvers = [r[0] for r in sorted_resolvers if not r[2] or r[2] == local_node_uuid]
        # remove duplicate resolver names but keeping the order
        seen = set()
        return [x for x in resolvers if not (x in seen or seen.add(x))]

    def _get_resolvers(self, all_resolvers=False):
        """
        This returns the list of the resolvernames of the user.
        If no resolver attribute exists at the moment, the user is searched
        in the realm and according to this the resolver attribute is set.

        It will only return one resolver in the list for backward compatibility

        .. note:: If the user does not exist in the realm, then an empty
           list is returned!

        :param all_resolvers: return all resolvers (of a realm), in which
            the user is contained
        :return: list of resolvers for self.login
        :rtype: list of strings
        """
        if self.resolver:
            return [self.resolver]

        resolvers = []
        for resolvername in self.get_ordered_resolvers():
            # test, if the user is contained in this resolver
            if self._locate_user_in_resolver(resolvername):
                break
        if self.resolver:
            resolvers = [self.resolver]
        return resolvers

    def _locate_user_in_resolver(self, resolvername):
        """
        Try to locate the user (by self.login) in the resolver with the given name.
        In case of success, this sets `self.resolver` as well as `self.uid`
        and returns True. If the resolver does not exist or the user does
        not exist in the resolver, False is returned.
        :param resolvername: string denoting the resolver name
        :return: boolean
        """
        y = get_resolver_object(resolvername)
        if y is None:  # pragma: no cover
            log.info("Resolver {0!r} not found!".format(resolvername))
            return False
        else:
            uid = y.getUserId(self.login)
            if uid not in ["", None]:
                log.info("user {0!r} found in resolver {1!r}".format(self.login,
                                                                     resolvername))
                log.info("userid resolved to {0!r} ".format(uid))
                self.resolver = resolvername
                self.uid = uid
                # We do not need to search other resolvers!
                return True
            else:
                log.debug("user {0!r} not found"
                          " in resolver {1!r}".format(self.login, resolvername))
                return False

    def get_user_identifiers(self):
        """
        This returns the UserId  information from the resolver object and
        the resolvertype and the resolvername
        (former: getUserId)
        (former: getUserResolverId)
        :return: The userid, the resolver type and the resolver name
                 like (1000, "passwdresolver", "resolver1")
        :rtype: tuple
        """
        if not self.resolver:
            raise UserError("The user can not be found in any resolver in "
                            "this realm!")
        return self.uid, self.rtype, self.resolver

    def exist(self):
        """
        Check if the user object exists in the user store
        :return: True or False
        """
        return bool(self.uid)

    @property
    def info(self):
        """
        return the detailed information for the user

        :return: a dict with all the userinformation
        :rtype: dict
        """
        if self.is_empty():
            # An empty user has no info
            return {}
        (uid, _rtype, _resolver) = self.get_user_identifiers()
        if uid is None:
            return {}
        y = get_resolver_object(self.resolver)
        user_info = y.getUserInfo(uid)
        # Now add the custom attributes, this is used e.g. in ADDUSERINRESPONSE
        user_info.update(self.attributes)
        return user_info

    @log_with(log)
    def set_attribute(self, attrkey, attrvalue, attrtype=None):
        """
        Set a custom attribute for a user

        :param attrkey: The key of the attribute
        :param attrvalue: The value of the attribute
        :return: The id of the attribute setting
        """
        ua = CustomUserAttribute(user_id=self.uid, resolver=self.resolver, realm_id=self.realm_id,
                                 Key=attrkey, Value=attrvalue, Type=attrtype).save()
        return ua

    @property
    def attributes(self):
        """
        returns the custom attributes of a user
        :return: a dictionary of attributes with keys and values
        """
        return get_attributes(self.uid, self.resolver, self.realm_id)

    @log_with(log)
    def delete_attribute(self, attrkey=None):
        """
        Delete the given key as custom user attribute.
        If no key is given, then all attributes are deleted

        :param attrkey: The key to delete
        :return: The number of deleted rows
        """
        if attrkey:
            ua = CustomUserAttribute.query.filter_by(user_id=self.uid, resolver=self.resolver,
                                                     realm_id=self.realm_id, Key=attrkey).delete()
        else:
            ua = CustomUserAttribute.query.filter_by(user_id=self.uid, resolver=self.resolver,
                                                     realm_id=self.realm_id).delete()
        db.session.commit()
        return ua

    @log_with(log)
    def get_user_phone(self, phone_type='phone', index=None):
        """
        Returns the phone number or a list of phone numbers of a user.

        :param phone_type: The type of the phone, i.e. either mobile or
                           phone (land line)
        :type phone_type: string
        :param index: The index of the selected phone number of list of the phones of the user.
            If the index is given, this phone number as string is returned.
            If the index is omitted, all phone numbers are returned.

        :returns: list with phone numbers of this user object
        """
        userinfo = self.info
        if phone_type in userinfo:
            phone = userinfo[phone_type]
            log.debug("got user phone {0!r} of type {1!r}".format(phone, phone_type))
            if type(phone) == list and index is not None:
                if len(phone) > index:
                    return phone[index]
                else:
                    log.warning("userobject ({0!r}) has not that much "
                                "phone numbers ({1!r} of {2!r}).".format(self, index, phone))
                    return ""
            else:
                return phone
        else:
            log.warning("userobject ({0!r}) has no phone of type {1!r}.".format(self, phone_type))
            return ""

    @log_with(log)
    def get_user_realms(self):
        """
        Returns a list of the realms, a user belongs to.
        Usually this will only be one realm.
        But if the user object has no realm but only a resolver,
        than all realms, containing this resolver are returned.
        This function is used for the policy module

        :return: realms of the user
        :rtype: list
        """
        allRealms = get_realms()
        Realms = []
        if self.realm == "" and self.resolver == "":
            defRealm = get_default_realm().lower()
            Realms.append(defRealm)
            self.realm = defRealm
        elif self.realm != "":
            Realms.append(self.realm.lower())
        else:
            # User has no realm!
            # we have got a resolver and will get all realms
            # the resolver belongs to.
            for key, val in allRealms.items():
                log.debug("evaluating realm {0!r}: {1!r} ".format(key, val))
                for reso in val.get('resolver', []):
                    resoname = reso.get("name")
                    if resoname == self.resolver:
                        Realms.append(key.lower())
                        log.debug("added realm %r to Realms due to "
                                  "resolver %r" % (key, self.resolver))
        return Realms

    @log_with(log, log_entry=False)
    def check_password(self, password):
        """
        The password of the user is checked against the user source

        :param password: The clear text password
        :return: the username of the authenticated user.
                 If unsuccessful, returns None
        :rtype: string/None
        """
        success = None

        # The password hash is used to avoid multiple password checks at the
        # resolver. It is not persisted in any way and only stored in memory for
        # the duration of the request.
        password_hash = hashlib.sha3_512(password.encode("utf-8")).hexdigest()

        try:
            log.info(f"User {self.login} from realm {self.realm} tries to authenticate")
            # If the password was already checked, return the known result
            if password_hash in self._checked_passwords.keys():
                if self._checked_passwords[password_hash]:
                    success = f"{self.login}@{self.realm}"
                    log.debug(f"Successfully authenticated user {self} from request cache.")
                else:
                    log.info(f"User {self} failed to authenticate from request cache.")
                return success
            res = self._get_resolvers()
            # Now we know, the resolvers of this user, and we can verify the password
            if len(res) == 1:
                y = get_resolver_object(self.resolver)
                uid, _rtype, _rname = self.get_user_identifiers()
                if y.checkPass(uid, password):
                    success = f"{self.login}@{self.realm}"
                    log.debug(f"Successfully authenticated user {self}.")
                    self._checked_passwords[password_hash] = True
                else:
                    log.info(f"User {self} failed to authenticate.")
                    self._checked_passwords[password_hash] = False

            elif not res:
                log.error(f"The user {self!r} exists in NO resolver.")
        except UserError as e:  # pragma: no cover
            log.error(f"Error while trying to verify the username: {e}")
        except Exception as e:  # pragma: no cover
            log.error(f"Error checking password within module {e}")
            log.debug(f"{traceback.format_exc()}")

        return success

    @log_with(log)
    def get_search_fields(self):
        """
        Return the valid search fields of a user.
        The search fields are defined in the UserIdResolver class.

        :return: searchFields with name (key) and type (value)
        :rtype: dict
        """
        searchFields = {}

        for reso in self._get_resolvers():
            # try to load the UserIdResolver Class
            try:
                y = get_resolver_object(reso)
                sf = y.getSearchFields()
                searchFields[reso] = sf

            except Exception as e:  # pragma: no cover
                log.warning("module {0!r}: {1!r}".format(reso, e))

        return searchFields

    # If passwords should not be logged, we hide it from the log entry
    @log_with(log, hide_kwargs=["password"])
    def update_user_info(self, attributes, password=None):
        """
        This updates the given attributes of a user.
        The attributes can be "username", "surname", "givenname", "email",
        "mobile", "phone", "password"

        :param attributes: A dictionary of the attributes to be updated
        :type attributes: dict
        :param password: The password of the user
        :return: True in case of success
        """
        if password is not None:
            attributes["password"] = password
        success = False
        try:
            log.info("User info for user {0!r}@{1!r} about to "
                     "be updated.".format(self.login, self.realm))
            res = self._get_resolvers()
            # Now we know, the resolvers of this user and we can update the
            # user
            if len(res) == 1:
                y = get_resolver_object(self.resolver)
                if not y.updateable:  # pragma: no cover
                    log.warning("The resolver {0!r} is not updateable.".format(y))
                else:
                    uid, _rtype, _rname = self.get_user_identifiers()
                    if y.update_user(uid, attributes):
                        success = True
                        # Delete entries corresponding to the old username from the user cache
                        delete_user_cache(username=self.login, resolver=self.resolver)
                        # If necessary, update the username
                        if attributes.get("username"):
                            self.login = attributes.get("username")
                        log.info("Successfully updated user {0!r}.".format(self))
                    else:  # pragma: no cover
                        log.info("user {0!r} failed to update.".format(self))

            elif not res:  # pragma: no cover
                log.error("The user {0!r} exists in NO resolver.".format(self))
        except UserError as exx:  # pragma: no cover
            log.error("Error while trying to verify the username: {0!r}".format(exx))

        return success

    @log_with(log)
    def delete(self):
        """
        This deletes the user in the user store. I.e. the user in the SQL
        database or the LDAP gets deleted.

        Returns True in case of success
        """
        success = False
        try:
            log.info("User {0!r}@{1!r} about to be deleted.".format(self.login, self.realm))
            res = self._get_resolvers()
            # Now we know, the resolvers of this user and we can delete it
            if len(res) == 1:
                y = get_resolver_object(self.resolver)
                if not y.updateable:  # pragma: no cover
                    log.warning("The resolver {0!r} is not updateable.".format(y))
                else:
                    uid, _rtype, _rname = self.get_user_identifiers()
                    if y.delete_user(uid):
                        success = True
                        log.info("Successfully deleted user {0!r}.".format(self))
                        # Delete corresponding entry from the user cache
                        delete_user_cache(username=self.login, resolver=self.resolver)
                    else:  # pragma: no cover
                        log.info("user {0!r} failed to update.".format(self))

            elif not res:  # pragma: no cover
                log.error("The user {0!r} exists in NO resolver.".format(self))
        except UserError as exx:  # pragma: no cover
            log.error("Error while trying to verify the username: {0!r}".format(exx))
        except Exception as exx:  # pragma: no cover
            log.error("Error checking password within module {0!r}".format(exx))
            log.debug("{0!s}".format(traceback.format_exc()))

        return success


@log_with(log, hide_kwargs=["password"])
def create_user(resolvername, attributes, password=None):
    """
    This creates a new user in the given resolver. The resolver must be
    editable to do so.

    The attributes is a dictionary containing the keys "username", "email",
    "phone",
    "mobile", "surname", "givenname", "password".

    We return the UID and not the user object, since the user could be located
    in several realms!

    :param resolvername: The name of the resolver, in which the user should
        be created
    :type resolvername: basestring
    :param attributes: Attributes of the user
    :type attributes: dict
    :param password: The password of the user
    :return: The uid of the user object
    """
    if password is not None:
        attributes["password"] = password
    y = get_resolver_object(resolvername)
    uid = y.add_user(attributes)
    return uid


@log_with(log)
def split_user(username):
    """
    Split the username of the form user@realm into the username and the realm
    splitting myemail@emailprovider.com@realm is also possible and will
    return (myemail@emailprovider.com, realm).

    If for a user@domain the "domain" does not exist as realm, the name is
    not split, since it might be the user@domain in the default realm

    If the Split@Sign configuration is disabled, the username won't be split
    and the username and an empty realm will be returned.

    We can also split realm\\user to (user, realm)

    :param username: the username to split
    :type username: string
    :return: username and realm
    :rtype: tuple
    """
    user = username.strip()
    realm = ""

    split_at_sign = get_from_config(SYSCONF.SPLITATSIGN, return_bool=True)
    if split_at_sign:
        l = user.split('@')
        if len(l) >= 2:
            if realm_is_defined(l[-1]):
                # split the last only if the last part is really a realm
                (user, realm) = user.rsplit('@', 1)
        else:
            l = user.split('\\')
            if len(l) >= 2:
                (realm, user) = user.rsplit('\\', 1)

    return user, realm


@log_with(log, hide_args_keywords={0: ["pass", "password"]})
def get_user_from_param(param, optionalOrRequired=optional):
    """
    Find the parameter user, realm and resolver and
    create a user object from these parameters.

    An exception is raised, if a user in a realm is found in more
    than one resolver.

    :param param: The dictionary of request parameters
    :type param: dict
    :param optionalOrRequired: whether the user is required
    :type optionalOrRequired: bool
    :return: User as found in the parameters
    :rtype: User object
    """
    realm = ""
    username = getParam(param, "user", optionalOrRequired)

    if username is None:
        username = ""
    else:
        username, realm = split_user(username)

    if "realm" in param:
        realm = param["realm"]

    if username != "":
        if realm is None or realm == "":
            realm = get_default_realm()

    user_object = User(login=username, realm=realm,
                       resolver=param.get("resolver"))

    return user_object


@log_with(log)
def get_user_list(param=None, user=None, custom_attributes=False):
    """
    This function returns a list of user dictionaries.

    :param param: search parameters
    :type param: dict
    :param user:  a specific user object to return
    :type user: User object
    :param custom_attributes:  Set to True, if you want to receive custom attributes
        of external users.
    :type custom_attributes: bool
    :return: list of dictionaries
    """
    users = []
    resolvers = []
    searchDict = {"username": "*"}
    param = param or {}

    # we have to recreate a new searchdict without the realm key
    # as delete does not work
    for key in param:
        lval = param[key]
        if key in ["realm", "resolver", "user", "username"]:
            continue
        searchDict[key] = lval
        log.debug("Parameter key:{0!r}={1!r}".format(key, lval))

    # update searchdict depending on existence of 'user' or 'username' in param
    # Since 'user' takes precedence over 'username' we have to check the order
    if 'username' in param:
        searchDict['username'] = param['username']
    if 'user' in param:
        searchDict['username'] = param['user']
    log.debug('Changed search key to username: %s.', searchDict['username'])

    # determine which scope we want to show
    param_resolver = getParam(param, "resolver")
    param_realm = getParam(param, "realm")
    user_resolver = None
    user_realm = None
    if user is not None:
        user_resolver = user.resolver
        user_realm = user.realm

    # Append all possible resolvers
    if param_resolver:
        resolvers.append(param_resolver)
    if user_resolver:
        resolvers.append(user_resolver)

    local_node_uuid = get_app_config_value("PI_NODE_UUID")
    for pu_realm in [param_realm, user_realm]:
        if pu_realm:
            realm_config = get_realm(pu_realm)
            for r in realm_config.get("resolver", {}):
                if r.get("name"):
                    if not r.get("node") or r["node"] == local_node_uuid:
                        resolvers.append(r.get("name"))

    if not (param_resolver or user_resolver or param_realm or user_realm):
        # if no realm or resolver was specified, we search the resolvers
        # in all realms
        all_realms = get_realms()
        for _name, res_list in all_realms.items():
            for resolver_entry in res_list.get("resolver"):
                if not resolver_entry.get("node") or resolver_entry["node"] == local_node_uuid:
                    resolvers.append(resolver_entry.get("name"))

    for resolver_name in set(resolvers):
        try:
            log.debug("Check for resolver class: {0!r}".format(resolver_name))
            y = get_resolver_object(resolver_name)
            log.debug("with this search dictionary: {0!r} ".format(searchDict))
            ulist = y.getUserList(searchDict)
            # Add resolvername to the list
            realm_id = get_realm_id(param_realm or user_realm)
            for ue in ulist:
                ue["resolver"] = resolver_name
                ue["editable"] = y.editable
            if custom_attributes and realm_id is not None:
                for ue in ulist:
                    # Add the custom attributes, by class method from User
                    # with uid, resolvername and realm_id, which we need to determine by the realm name
                    ue.update(get_attributes(ue.get("userid"), ue.get("resolver"), realm_id))
            log.debug("Found this userlist: {0!r}".format(ulist))
            users.extend(ulist)

        except KeyError as exx:  # pragma: no cover
            log.error("{0!r}".format(exx))
            log.debug("{0!s}".format(traceback.format_exc()))
            raise exx

        except Exception as ex:  # pragma: no cover
            log.error(f"Unable to get user list: {ex}")
            log.debug("{0!s}".format(traceback.format_exc()))
            continue

    return users


@log_with(log)
@user_cache(cache_username)
def get_username(userid, resolvername):
    """
    Determine the username for a given id and a resolvername.

    :param userid: The id of the user in a resolver
    :type userid: string
    :param resolvername: The name of the resolver
    :return: the username or "" if it does not exist
    :rtype: string
    """
    username = ""
    if userid:
        y = get_resolver_object(resolvername)
        if y:
            username = y.getUsername(userid)
    return username


def log_used_user(user: User, other_text: str = "") -> str:
    """
    This creates a log message combined of a user and another text.
    The user information is only added, if user.login != user.used_login

    :param user: A user to log
    :type user:  User object
    :param other_text: Some additional text
    :return: str
    """
    return "logged in as {0}. {1}".format(user.used_login, other_text) if user.used_login != user.login else other_text


def get_attributes(uid, resolver, realm_id):
    """
    Returns the attributes for the given user.

    :param uid: The UID of the user
    :param resolver: The name of the resolver
    :param realm_id: The realm_id
    :return: A dictionary of key/values
    """
    r = {}
    attributes = CustomUserAttribute.query.filter_by(user_id=uid, resolver=resolver, realm_id=realm_id).all()
    for attr in attributes:
        r[attr.Key] = attr.Value
    return r


def is_attribute_at_all():
    """
    Check if there are custom user attributes at all
    :return: bool
    """
    return bool(CustomUserAttribute.query.count())



#  2018-05-09 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             TAN token with to be randomly used TAN values
#
#  (c) 2018 Cornelius Kölbel - cornelius.koelbel@netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This file contains the definition of the tan token class
It depends on the DB model, and the lib.tokenclass.
"""

import logging

from privacyidea.lib.decorators import check_token_otp_length
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.tokens.papertoken import PaperTokenClass
from privacyidea.lib.policy import SCOPE, ACTION, GROUP
from privacyidea.lib import _
from privacyidea.lib.crypto import geturandom, hash

log = logging.getLogger(__name__)
DEFAULT_COUNT = 100
SALT_LENGTH = 4


class TANACTION(object):
    TANTOKEN_COUNT = "tantoken_count"


class TanTokenClass(PaperTokenClass):
    """
    The TAN token allows to print out the next e.g. 100 OTP values.
    This sheet of paper can be used to authenticate and strike out the used
    OTP values. In contrast to the paper token, the OTP values of the TAN token
    can be used randomly.
    """

    @log_with(log)
    def __init__(self, db_token):
        """
        This creates a new TAN token object from a DB token object.

        :param db_token: instance of the orm db object
        :type db_token:  orm object
        """
        TokenClass.__init__(self, db_token)
        self.set_type("tan")
        self.hKeyRequired = False

    @staticmethod
    def get_class_type():
        """
        return the token type shortname

        :return: 'paper'
        :rtype: string
        """
        return "tan"

    @staticmethod
    def get_class_prefix():
        """
        Return the prefix, that is used as a prefix for the serial numbers.
        :return: PITN
        """
        return "PITN"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict or scalar
        """
        res = {'type': 'tan',
               'title': 'TAN Token',
               'description': _('TAN: TANs printed on a sheet of paper.'),
               'init': {},
               'config': {},
               'user':  ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {
                   SCOPE.ENROLL: {
                       TANACTION.TANTOKEN_COUNT: {
                           "type": "int",
                           "desc": _("The number of OTP values, which are "
                                     "printed on the paper.")
                       },
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of TAN tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _(
                               "The user may only have this maximum number of active TAN tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               }
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res
        return ret

    def update(self, param, reset_failcount=True):
        if "tans" in param:
            # init tokens with tans
            tans = param.get("tans").split()
            tan_dict = {k: v for k, v in enumerate(tans)}
            # Avoid to generate TANs in the superclass PaperToken, since we get the tans from params
            param["papertoken_count"] = 0
            # Determine the otplen from the TANs
            if len(tans) > 0:
                param["otplen"] = len(tans[0])
            PaperTokenClass.update(self, param, reset_failcount=reset_failcount)
        else:
            # Init token without tans, so we create tans in the superclass PaperToken
            param["papertoken_count"] = param.get("tantoken_count") or DEFAULT_COUNT
            PaperTokenClass.update(self, param, reset_failcount=reset_failcount)
            # After this creation, the init_details contain the complete list of the TANs
            tan_dict = self.init_details.get("otps", {})
        for tankey, tanvalue in tan_dict.items():
            # Get a 4 byte salt from the crypto module
            salt = geturandom(SALT_LENGTH, hex=True)
            # Now we add all TANs to the tokeninfo of this token.
            hashed_tan = hash(tanvalue, salt)
            self.add_tokeninfo("tan.tan{0!s}".format(tankey),
                               "{0}:{1}".format(salt, hashed_tan))

    @check_token_otp_length
    def check_otp(self, anOtpVal, counter=None, window=None, options=None):
        """
        check if the given OTP value is valid for this token.

        :param anOtpVal: the to be verified otpvalue
        :type anOtpVal: string
        :param counter: the counter state, that should be verified
        :type counter: int
        :param window: the counter +window, which should be checked
        :type window: int
        :param options: the dict, which could contain token specific info
        :type options: dict
        :return: the counter state or -1
        :rtype: int
        """
        res = -1
        tans = self.get_tokeninfo()
        for tankey, tanvalue in tans.items():
            if tankey.startswith("tan.tan"):
                salt, tan = tanvalue.split(":")
                if tan == hash(anOtpVal, salt):
                    self.del_tokeninfo(tankey)
                    return 1

        return res

    @staticmethod
    def get_import_csv(l):
        """
        Read the list from a csv file and return a dictionary, that can be used
        to do a token_init.

        :param l: The list of the line of a csv file
        :type l: list
        :return: A dictionary of init params
        """
        params = TokenClass.get_import_csv(l)
        # Delete the otplen, if it exists. The fourth column is the TANs!
        if "otplen" in params:
            del params["otplen"]

        # tans
        if len(l) >= 4:
            params["tans"] = l[3]

        return params

    def get_as_dict(self):
        """
        This returns the token data as a dictionary.
        It is used to display the token list at /token/list.

        The TAN token class removes the tan.tanXX information and
        only returns the number of remaining tans.

        :return: The token data as dict
        :rtype: dict
        """
        # first get the database values as dict
        token_dict = self.token.get()

        if "info" in token_dict:
            tan_count = 0
            filtered_info = {}
            for infokey, infovalue in token_dict["info"].items():
                if infokey.startswith("tan.tan"):
                    tan_count += 1
                else:
                    filtered_info[infokey] = infovalue
            filtered_info["tan.count"] = tan_count
            token_dict["info"] = filtered_info

        return token_dict
